# Copyright (c) 2020, Anthony Nouy, Erwan Grelier
# This file is part of tensap (tensor approximation package).

# tensap is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# tensap is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with tensap.  If not, see <https://www.gnu.org/licenses/>.

'''
Module full_tensor.

'''

# from string import ascii_lowercase
import numpy as np
import tensap


class FullTensor:
    '''
    Class FullTensor.

    Attributes
    ----------
    data : numpy.ndarray
        The content of the tensor.
    order : int
        The order of the tensor.
    shape : numpy.ndarray
        The shape of the tensor.
    is_orth : bool
        Boolean indicating if the representation of the tensor is orthogonal
        (i.e. one mu-matricization is orthogonal).
    orth_dim : bool
        Boolean indicating, if is_orth = True, the dimension mu for which
        the mu-matricization of the tensor is orthogonal.

    '''

    # Override numpy's operations with reversed operands
    __array_priority__ = 1

    def __init__(self, data, order=None, shape=None):
        '''
        Constructor for the class FullTensor.

        Parameters
        ----------
        data : numpy.ndarray or tensap.FullTensor or tensorflow.Tensor
            The content of the tensor.
        order : int, optional
            The order of the tensor. The default is None, indicating to infer
            it from data.
        shape : list or numpy.ndarray, optional
            The shape of the tensor. The default is None, the shape is deduced
            from data.

        Returns
        -------
        None.

        '''
        if hasattr(data, 'data'):
            self.data = np.array(data.data)
        else:
            self.data = np.array(data)
        if shape is not None:
            self.data = np.reshape(self.data, shape, order='F')

        ndim = np.ndim(self.data)
        if order is not None and ndim != order:
            for d in np.arange(ndim, order):
                self.data = np.expand_dims(self.data, d)

        self.is_orth = False
        self.orth_dim = None

    @property
    def order(self):
        '''
        Compute the order of the tensor.

        Returns
        -------
        int
            The order of the tensor.

        '''
        return np.ndim(self.data)

    @property
    def ndim(self):
        '''
        Compute the order of the tensor. Equivalent to self.order.

        Returns
        -------
        int
            The order of the tensor.

        '''
        return self.data.ndim

    @property
    def size(self):
        '''
        Compute the number of elements of the tensor.

        Returns
        -------
        numpy.ndarray
            The number of elements of the tensor.

        '''
        return np.array(self.data.shape)

    @property
    def shape(self):
        '''
        Compute the shape of the tensor

        Returns
        -------
        numpy.ndarray
            The shape of the tensor.

        '''
        return np.array(self.data.shape)

    def tree_based_tensor(self):
        '''
        Convert a FullTensor into a TreeBasedTensor.

        Returns
        -------
        TreeBasedTensor
            The FullTensor in tree-based tensor format.

        '''
        tree = tensap.DimensionTree.trivial(self.order)
        tensors = [FullTensor(self)]
        for dim in range(self.order):
            tensors.append(FullTensor(np.eye(self.shape[dim])))
        return tensap.TreeBasedTensor(tensors, tree)

    def sparse(self):
        '''
        Conversion of a FullTensor into a SparseTensor.

        Returns
        -------
        tensap.SparseTensor
            A SparseTensor representation of the FullTensor.

        '''
        dat = np.reshape(self.data, -1, order='F')
        ind = np.nonzero(dat)[0]
        indices = tensap.MultiIndices.ind2sub(self.shape, ind)
        return tensap.SparseTensor(dat, indices, self.shape)

    def __repr__(self):
        return ('<{} FullTensor:{n}' +
                '{t}order = {},{n}' +
                '{t}shape = {},{n}' +
                '{t}is_orth = {},{n}' +
                '{t}orth_dim = {}>').format('x'.join(map(str, self.shape)),
                                            self.order,
                                            self.shape,
                                            self.is_orth,
                                            self.orth_dim,
                                            t='\t', n='\n')

    def __getitem__(self, key):
        return self.data.__getitem__(key)

    def __eq__(self, tensor2):
        return np.all(self.data == tensor2.data)

    def __add__(self, arg):
        if isinstance(arg, FullTensor):
            arg = arg.data
        return FullTensor(self.data + arg)

    def __radd__(self, arg):
        return self + arg

    def __sub__(self, arg):
        if isinstance(arg, FullTensor):
            arg = arg.data
        return FullTensor(self.data - arg)

    def __rsub__(self, arg):
        return -self + arg

    def __neg__(self):
        return FullTensor(-self.data)

    def __mul__(self, arg):
        if isinstance(arg, FullTensor):
            arg = arg.data
        return FullTensor(self.data * arg)

    def __rmul__(self, arg):
        return self * arg

    def __truediv__(self, arg):
        if isinstance(arg, FullTensor):
            arg = arg.data
        return FullTensor(self.data / arg)

    def __pow__(self, arg):
        assert np.isscalar(arg), 'The power must be a scalar.'
        return FullTensor(self.data ** arg)

    def hadamard_product(self, arg):
        '''
        Compute the Hadamard product of two tensors.

        Equivalent to self * arg.

        Parameters
        ----------
        arg : tensap.FullTensor or numpy.ndarray
            The second tensor of the Hadamard product.

        Returns
        -------
        FullTensor
            The tensor resulting from the Hadamard product.

        '''
        return self * arg

    def storage(self):
        '''
        Return the storage complexity of the FullTensor.

        Returns
        -------
        int
            The storage complexity of the FullTensor.

        '''
        return np.size(self.data)

    def sparse_storage(self):
        '''
        Return the sparse storage complexity of the FullTensor.

        Returns
        -------
        int
            The sparse storage complexity of the FullTensor.

        '''
        return np.count_nonzero(self.data)

    def numpy(self):
        '''
        Convert the FullTensor to a numpy.ndarray.

        Returns
        -------
        numpy.ndarray
            The FullTensor as a numpy.ndarray.

        '''
        return self.data

    def eval_at_indices(self, indices, dims=None):
        '''
        Evaluate the tensor at indices.

        If dims is None, return
        s(k) = x(indices(k, 1), indices(k, 2), ..., indices(k, d)),
        1 <= k <= self.shape[0].

        If dims is not None, return a partial evaluation: up to a permutation
        (placing the dimensions dims on the left), return
        s(k, i_1, ..., i_d') = x(indices(k, 1), indices(k, 2), ...,
        indices(k, M), i_1, ..., i_d'),
        1 <= k <= self.shape[0], with M = dims.size and d' = self.order - M.

        Parameters
        ----------
        indices : list of numpy.ndarray
            The indices of the tensor.
        dims : list of numpy.ndarray, optional
            The dimensions associated with the indices. The default is None,
            indicating that indices refers to all the dimensions.

        Returns
        -------
        evaluations : numpy.ndarray or FullTensor
            The evaluations of the tensor.

        '''
        indices = np.atleast_2d(indices)
        if dims is None:
            dims = np.arange(self.order)
        else:
            dims = np.atleast_1d(dims)
            if indices.shape[1] != dims.size:
                indices = np.transpose(indices)
            assert dims.size == indices.shape[1], \
                'Wrong size of multi-indices.'
            sort_ind = np.argsort(dims)
            dims = dims[sort_ind]
            indices = indices[:, sort_ind]
        assert dims.size == indices.shape[1], 'Wrong size of multi-indices.'

        if dims.size == self.order:
            data = self
            evaluations = np.array([data[tuple(i)] for i in indices.tolist()])
        elif dims.size == 1:
            ind = [':']*self.order
            ind[dims[0]] = np.ravel(indices).tolist()
            evaluations = self.sub_tensor(*ind)
        else:
            no_dims = tensap.fast_setdiff(np.arange(self.order), dims)
            indices = np.ravel_multi_index(np.transpose(indices),
                                           [self.shape[i] for i in dims])
            evaluations = self.matricize(dims).sub_tensor(indices, ':')
            evaluations = evaluations.reshape([indices.size] +
                                              [self.shape[i] for i in no_dims])
            left_dims = np.arange(dims[0])
            evaluations = evaluations.transpose(
                np.concatenate((np.arange(1, left_dims.size + 1), [0],
                                np.arange(left_dims.size + 1,
                                          self.order - dims.size + 1))))
        return evaluations

    def eval_diag(self, dims=None):
        '''
        Extract the diagonal of the tensor.

        The tensor must be such that self.shape[mu] = n for all mu (in dims if
        provided).

        Parameters
        ----------
        dims : list of numpy.ndarray, optional
            The dimensions associated with the indices of the diagonal. The
            default is None,indicating that the indices refer to all the
            dimensions.

        Returns
        -------
        data : numpy.ndarray
            The evaluations of the diagonal of the tensor.

        '''
        if dims is None:
            dims = np.arange(self.order)
        else:
            dims = np.atleast_1d(dims)

        if dims.size == 1:
            data = self
        else:
            assert np.all([self.shape[x] == self.shape[dims[0]] for
                           x in dims]),\
             'The shapes of the tensor in dimensions dims should be equal.'
            ind = np.repeat(np.reshape(np.arange(self.shape[dims[0]]),
                                       [-1, 1]), dims.size, 1)
            data = self.eval_at_indices(ind, dims)
        return data

    def reshape(self, shape):
        '''
        Reshape the tensor.

        Parameters
        ----------
        shape : list or numpy.ndarray
            The new shape of the tensor.

        Returns
        -------
        tensor : FullTensor
            The reshaped tensor.

        '''
        tensor = FullTensor(self)
        tensor.data = np.reshape(tensor.data, shape, order='F')
        return tensor

    def transpose(self, dims):
        '''
        Transpose (permute) the dimensions of the tensor.

        Parameters
        ----------
        dims : list or numpy.ndarray
            The new ordering of the dimensions.

        Returns
        -------
        tensor : FullTensor
            The transposed (permuted) tensor.

        '''
        tensor = FullTensor(self)
        tensor.data = np.transpose(tensor.data, dims)
        return tensor

    def itranspose(self, dims):
        '''
        Return the inverse transpose (permutation) of the dimensions of the
        tensor.

        Parameters
        ----------
        dims : list or numpy.ndarray
            The original transpose (permutation) indices.

        Returns
        -------
        FullTensor
            The transposed (permuted) tensor.

        '''
        return self.transpose(np.argsort(dims))

    def squeeze(self, dims=None):
        '''
        Remove the singleton dimensions of the tensor.

        Parameters
        ----------
        dims : list or numpy.ndarray, optional
            Dimensions to squeeze. The default is None, indicating all the
            singleton dimensions.

        Returns
        -------
        out : float or FullTensor
            The squeezed tensor.

        '''
        if dims is not None:
            dims = tuple(dims)

        out = FullTensor(np.squeeze(self.data, dims))
        if out.order == 0:
            out = out.data
        return out

    def dot(self, tensor2):
        '''
        Return the inner product of two tensors.

        Parameters
        ----------
        tensor2 : FullTensor
            The second tensor of the inner products.

        Returns
        -------
        numpy.float
            The inner product of the two tensors.

        '''
        return np.sum(np.multiply(self.data, tensor2.data))

    def norm(self):
        '''
        Compute the canonical norm of the FullTensor.

        Returns
        -------
        numpy.float
            The norm of the tensor.

        '''
        return np.linalg.norm(self.data)

    def full(self):
        '''
        Return the tensor.

        Returns
        -------
        FullTensor
            The tensor.

        '''
        return self

    def sub_tensor(self, *indices):
        '''
        Extract a subtensor of the tensor.

        The result is a tensor s of shape
        len(indices[0]), ..., len(indices[self.order-1]),
        such that
        s(k1,...,kd) = x(indices[0][k1], ..., indices[self.order-1][kd]).

        Example: x.subTensor([1, 2], ':', [2, 5, 6]) returns a tensor with
        shape [2, self.shape[1], 3].

        Parameters
        ----------
        *indices : list
            The indices to extract in each dimension. ':' indicates all the
            indices.

        Returns
        -------
        FullTensor
            The subtensor.

        '''
        data = self.data
        order = self.order
        for dim in range(self.order):
            ind_loc = np.atleast_1d(indices[dim])
            if ind_loc.size != 1 or ind_loc[0] != ':':
                data = np.take(data, ind_loc, axis=dim)
                if order != data.ndim:
                    data = np.expand_dims(data, dim)
                order = data.ndim
        return FullTensor(data)

    def cat(self, tensor2, dims=None):
        '''
        Concatenate the tensors.

        Concatenates self and tensor2 in a tensor z such that:
        z(i_1 ,..., i_d) = x(i_1, ..., i_d) if i_k <= sz[k]-1 for k in dims,
        z(i_1, ..., i_d) = y(i_1-sz[0], ..., i_d-sz[d-1]) if i_k >= sz[k]
        for k in dims,
        z(i_1, ..., i_d) = 0 otherwise, with sz = self.shape and
        dims = range(self.order) if not provided.

        Parameters
        ----------
        tensor2 : FullTensor
            The second tensor to be concatenaed.
        dims : list or numpy.ndarray, optional
            The dimensions of the concatenation. The default is None,
            indicating all the dimensions.
        Returns
        -------
        data : FullTensor
            The concatenated tensors.

        '''
        assert self.order == tensor2.order, \
            'The orders of the tensors must be equal.'

        tensor1 = FullTensor(self)
        order = self.order
        shape1 = np.atleast_1d(self.shape)
        shape2 = np.atleast_1d(tensor2.shape)

        if dims is None:
            dims = range(self.order)

        dims = np.atleast_1d(dims)
        dims_not = tensap.fast_setdiff(np.arange(order), dims)
        assert np.all([a == b for a, b in zip(shape1[dims_not],
                                              shape2[dims_not])]), \
            'The dimensions of the tensors are not compatible.'

        if dims.size == 1:
            data = np.concatenate([tensor1.data, tensor2.data], dims[0])
        else:
            shape_out = np.array(shape1)
            shape_out[dims] = shape1[dims] + shape2[dims]

            padding = np.transpose([[0]*order, shape_out - shape1])
            data = np.pad(tensor1.data, padding)
            padding = np.transpose([shape_out - shape2, [0]*order])
            data += np.pad(tensor2.data, padding)

        return FullTensor(data)

    def reduce_sum(self, dims=None):
        '''
        Compute the sum of elements across dimensions dims of a tensor.

        Similar to tensorflow.reduce_sum.

        Parameters
        ----------
        dims : list or numpy.ndarray, optional
            The dimensions to be reduced. The default is None, indicating all
            the dimensions.

        Returns
        -------
        FullTensor
            The reduced tensor.

        '''
        return FullTensor(np.sum(self.data, dims, keepdims=True))

    def reduce_mean(self, dims=None):
        '''
        Compute the mean of elements across dimensions dims of a tensor.

        Similar to tensorflow.mean.

        Parameters
        ----------
        dims : list or numpy.ndarray, optional
            The dimensions to be reduced. The default is None, indicating all
            the dimensions.

        Returns
        -------
        FullTensor
            The reduced tensor.

        '''
        return FullTensor(np.mean(self.data, dims, keepdims=True))

    def kron(self, tensor2):
        '''
        Kronecker product of tensors.

        Similar to numpy.kron but for arbitrary tensors.

        Parameters
        ----------
        tensor2 : FullTensor
            The second tensor of the Kronecker product.

        Returns
        -------
        FullTensor
            The tensor resulting from the Kronecker product.

        '''
        order1 = self.order
        order2 = tensor2.order
        order_max = np.max((order1, order2))

        shape1 = np.concatenate((self.shape,
                                np.ones(order_max - order1, dtype=int)))
        shape2 = np.concatenate((tensor2.shape,
                                np.ones(order_max - order2, dtype=int)))

        data1 = np.reshape(self.data, [-1, 1])
        data2 = np.reshape(tensor2.data, [1, -1])

        perm = np.reshape(np.transpose(np.reshape(np.arange(2*order_max),
                                                  [2, order_max])),
                          2*order_max)

        data = np.reshape(np.transpose(np.reshape(np.matmul(data1, data2),
                                                  np.concatenate((shape1,
                                                                  shape2))),
                                       perm), shape1 * shape2)
        return FullTensor(data, shape=self.shape*tensor2.shape)

    def orth(self, dim=None):
        '''
        Orthogonalize the tensor.

        Parameters
        ----------
        dim : int, optional
            The dimension of the orthogonal dim-matricization of self. The
            default is None, returning a copy of the original tensor.

        Returns
        -------
        tensor : FullTensor
            A tensor whose dim-matricization is an orthogonal matrix
            corresponding to the Q factor of a QR factorization of the
            dim-matricization of self.
        r_matrix : numpy.ndarray
            The R factor.

        '''
        tensor = FullTensor(self)  # Copy the tensor

        if dim is None:
            return tensor, np.array([])

        if dim == -1:
            dim = tensor.order-1

        dims = np.concatenate((np.arange(dim),
                               np.arange(dim+1, tensor.order),
                               [dim]))
        tensor = tensor.transpose(dims)

        shape0 = np.array(tensor.shape)
        tensor = tensor.reshape([np.prod(shape0[:-1]), shape0[-1]])

        try:
            from tensorflow.python.ops.gen_linalg_ops import qr
            q_tf, r_tf = qr(tensor.data, full_matrices=False)
            tensor.data, r_matrix = q_tf.numpy(), r_tf.numpy()
        except ImportError:
            tensor.data, r_matrix = np.linalg.qr(tensor.data)

        shape0[-1] = r_matrix.shape[0]
        tensor = tensor.reshape(shape0)
        tensor = tensor.itranspose(dims)
        tensor.is_orth = True
        tensor.orth_dim = dim
        return tensor, r_matrix

    def dot_with_rank_one_metric(self, tensor2, matrix):
        '''
        Compute the weighted inner product of two tensors.

        Compute the weighted canonical inner product of self and tensor2,
        where the inner product related to dimension k is weighted by
        matrix[k]. It is equivalent to
        self.dot(tensor2.tensor_matrix_product(matrix)),
        but can be much faster.

        Parameters
        ----------
        tensor2 : FullTensor
            The second tensor of the inner product.
        matrix : list or numpy.ndarray or FullTensor
            The weight matrix.

        Returns
        -------
        numpy.float
            The weighted inner product.

        '''
        return self.dot(tensor2.tensor_matrix_product(matrix))

    def tensordot_matrix_product_except_dim(self, tensor2, matrices, dim):
        '''
        Particular type of contraction.

        Compute a special contraction of two tensors self, tensor2, a list of
        matrices matrices and a particular dimension dim. Note that dim must
        be a scalar, while matrices must be a list array with self.order
        elements.

        Parameters
        ----------
        tensor2 : FullTensor
            The second tensor of the contraction.
        matrices : list
            The list of matrices of the contraction.
        dim : int
            The excluded dimension.

        Returns
        -------
        numpy.ndarray
            The result of the contraction.

        '''
        assert isinstance(matrices, list), 'matrices should be a list.'
        assert len(matrices) == self.order, \
            'len(matrices) must be self.order.'

        dims = tensap.fast_setdiff(np.arange(self.order), dim)
        matrices = [matrices[i] for i in dims]
        tmp = tensor2.tensor_matrix_product(matrices, dims)
        tmp = self.tensordot(tmp, dims, dims)
        return tmp

    def tensordot(self, tensor2, dims1, dims2=None):
        '''
        Contract two tensors along specified dimensions.

        Similar to tensorflow.tensordot.

        Parameters
        ----------
        tensor2 : FullTensor
            The second tensor of the contraction.
        dims1 : list or int
            The dimensions of contractions for the first tensor.
        dims2 : list or int, optional
            The dimensions of contractions for the second tensor. The default
            is None which indicates, if dims1 = 0, to perform the outer
            product of the two tensors, similarly to tensorflow.tensordot.

        Returns
        -------
        out : FullTensor
            The resulting tensor.

        '''
        dims1 = np.atleast_1d(dims1)
        if dims1.size == 1 and dims1 == 0 and dims2 is None:
            # Outer product (notation similar to tensorflow)
            out = FullTensor(np.tensordot(self.data, tensor2.data, 0))
        else:
            dims2 = np.atleast_1d(dims2)
            assert np.all([self.shape[i] == tensor2.shape[j] for i, j in
                           zip(dims1, dims2)]), \
                'The dimensions of the tensors are not compatible.'
            out = FullTensor(np.tensordot(self.data, tensor2.data,
                                          [dims1, dims2]))
        return out

    def tensordot_eval_diag(self, tensor2, dims1, dims2, diag_dims1,
                            diag_dims2, diag=False):
        '''
        Evaluate of the diagonal of a tensor obtained by contraction of two
        tensors.

        The contraction is performed along the dimensions dims1 for self and
        dims2 for tensor2, and the diagonal is evaluated according to the
        dimensions diag_dims1 for self and diag_dims2 for tensor2.

        The boolean diag indicates if the several diagonals are evaluated, for
        instance:
        - if diag is False, for order-4 tensors x and y,
        z = x.tensordot_eval_diag(y,[1,3],[2,3],2,0) returns an order-3 tensor
        z(i1,k,j2) = sum_{l1,l2} x(i1,l1,k,l2) y(k,j2,l1,l2)
        - if diag is True, for order-5 tensors x and y,
        z = x.tensordot_eval_diag(y,[1,3],[2,3],[0,2],[1,4]) returns an
        order-4 tensor
        z(k,l,i5,j1) = sum_{l1,l2} x(k,l1,l,l2,i5) y(j1,k,l1,l2,l)

        Parameters
        ----------
        tensor2 : FullTensor
            The second tensor of the product.
        dims1 : list or numpy.ndarray
            Dimensions of the first tensor for the contraction.
        dims2 : list or numpy.ndarray
            Dimensions of the second tensor for the contraction.
        diag_dims1 : list or numpy.ndarray
            Indices of the first tensor for the evaluation of the diagonal.
        diag_dims2 : list or numpy.ndarray
            Indices of the second tensor for the evaluation of the diagonal.
        diag : bool, optional
            Boolean enabling the evaluation of multiple diagonals. The default
            is False.

        Returns
        -------
        FullTensor
            The evaluated tensor.

        '''
        # Check if an outer product is asked with dims1 and dims2 equal to None
        if dims1 is None and dims2 is None:
            dims1 = []
            dims2 = []

        dims1 = np.atleast_1d(dims1)
        dims2 = np.atleast_1d(dims2)
        diag_dims1 = np.atleast_1d(diag_dims1)
        diag_dims2 = np.atleast_1d(diag_dims2)

        ind1 = np.arange(self.ndim)
        ind2 = ind1.size + np.arange(tensor2.ndim)
        if diag and diag_dims1.size != 0 and diag_dims2.size != 0:
            ind2[diag_dims2] = ind1[diag_dims1]
        elif not diag:
            ind1[diag_dims1] = ind1[diag_dims1[0]]
            ind2[diag_dims2] = ind1[diag_dims1[0]]
        if dims1.size != 0 and dims2.size != 0:
            ind2[dims2] = ind1[dims1]
        ind12 = np.concatenate((ind1, ind2))
        indexes = np.unique(ind12, return_index=True)[1]
        # Retain only the unique values without sorting the vector
        ind_out = np.array([ind12[index] for index in sorted(indexes)])
        # Remove from ind_out the contracted dimensions without sorting it
        ind_out = ind_out[[i not in np.atleast_1d(dims1) for i in ind_out]]
        array = np.einsum(self.data, ind1.tolist(), tensor2.data,
                          ind2.tolist(), ind_out.tolist())
        # alph = list(ascii_lowercase)
        # ind1 = [alph[i] for i in ind1]
        # ind2 = [alph[i] for i in ind2]
        # ind_out = [alph[i] for i in ind_out]
        # array = np.einsum(''.join(ind1) + ', ' + ''.join(ind2) + ' -> ' +
        #                   ''.join(ind_out), self.data, tensor2.data)
        return FullTensor(array)

    def tensor_matrix_product(self, matrices, dims=None):
        '''
        Contract a tensor with matrices.

        The second dimension of the matrix matrices[k] is contracted with the
        k-th dimension of self, with the indices k given in dims (if provided).

        Parameters
        ----------
        matrices : numpy.ndarray or list of numpy.ndarray
            The matrices to use in the product.
        dims : list or numpy.ndarray, optional
            Indices of the contractions. The default is None, indicating all
            the dimensions.

        Returns
        -------
        FullTensor
            The tensor after the contractions with the matrices.

        '''
        if dims is None:
            assert isinstance(matrices, (list, np.ndarray)), \
                'matrices should be a list or a numpy.ndarray.'
            assert len(matrices) == self.order, \
                'len(matrices) must be self.order.'
            dims = range(self.order)
        else:
            dims = np.atleast_1d(dims)
            if not isinstance(matrices, list):
                matrices = [matrices]
            assert len(matrices) == dims.size, \
                'len(matrices) must be equal to dims.size.'

        # Numpy implementation
        # matrices = [np.array(x) for x in matrices]
        # data = self.numpy()
        # if self.order == 1:
        #     data = np.matmul(matrices[0], data)
        # else:
        #     k = 0
        #     for dim in np.nditer(dims):
        #         perm_dims = np.concatenate(([dim], np.arange(dim),
        #                                     np.arange(dim+1, self.order)))
        #         data = np.transpose(data, perm_dims)
        #         shape0 = np.array(data.shape)
        #         data = np.reshape(data, [shape0[0], np.prod(shape0[1:])])
        #         data = np.matmul(matrices[k], data)
        #         shape0[0] = matrices[k].shape[0]
        #         data = np.reshape(data, shape0)
        #         data = np.transpose(data, np.argsort(perm_dims))
        #         k += 1
        # return FullTensor(data)

        tensor = FullTensor(self)
        matrices = [FullTensor(x) for x in matrices]
        for i, dim in enumerate(dims):
            index = np.concatenate((
                tensap.fast_setdiff(np.arange(tensor.order), dim), [dim]))
            tensor = tensor.tensordot(matrices[i], dim, 1).itranspose(index)
        return tensor

    def tensor_vector_product(self, vectors, dims=None):
        '''
        Compute the contraction of the tensor with vectors.

        Compute the contraction of self with each vector contained in the list
        vectors along dimensions specified by dims. The operation is such that
        V[k] is contracted with the dims[k]-th dimension of self.

        Parameters
        ----------
        vectors : numpy.ndarray or list of numpy.ndarray
            The vectors to use in the product.
        dims : list or numpy.ndarray, optional
            Indices of the contractions. The default is None, indicating all
            the dimensions.

        Returns
        -------
        FullTensor
            The tensor after the contractions with the vectors.

        '''
        if dims is None:
            assert isinstance(vectors, list), 'vectors should be a list.'
            assert len(vectors) == self.order, \
                'len(vectors) must be self.order.'
            dims = np.arange(self.order)
        else:
            dims = np.array(dims)
            if not isinstance(vectors, list):
                vectors = [vectors]
            assert len(vectors) == dims.size, \
                'len(vectors) must be equal to dims.size.'

        vectors = [FullTensor(x, 2, [1, -1]) for x in vectors]
        return self.tensor_matrix_product(vectors, dims).squeeze(dims.tolist())

    def tensor_matrix_product_eval_diag(self, matrices, dims=None):
        '''
        Evaluate the diagonal of a tensor obtained by contraction with
        matrices.

        Provides the diagonal of the tensor obtained by contracting the tensor
        with matrices H[k] along dimensions dims(k)+1, for k = 0, ...,
        dims.size-1.

        Parameters
        ----------
        matrices : list
            The matrices to use in the product.
        dims : list or numpy.ndarray, optional
            Indices of the contractions. The default is None, indicating all
            the dimensions.

        Returns
        -------
        out : FullTensor
            The diagonal of the contractions of the tensor with the matrices.

        '''
        if dims is None:
            assert isinstance(matrices, list), 'matrices should be a list.'
            assert len(matrices) == self.order, \
                'len(matrices) must be self.order.'
            dims = np.arange(self.order)
        else:
            dims = np.atleast_1d(dims)
            if not isinstance(matrices, list):
                matrices = [matrices]
            assert len(matrices) == dims.size, \
                'len(matrices) must be equal to dims.size.'

        matrices = [FullTensor(x) for x in matrices]
        ind = np.flip(np.argsort(dims))
        out = matrices[ind[0]].tensordot(self, 1, dims[ind[0]])
        for i in ind[1:]:
            out = matrices[i].tensordot_eval_diag(out, 1, dims[i]+1, 0, 0)

        # if out.order == 1:
        #     out = out.numpy()
        return out

    def tensor_diagonal_matrix_product(self, matrices, dims=None):
        '''
        Contract a FullTensor with matrices built from their diagonals.

        The second dimension of the matrix matrices[k] is contracted with the
        k-th dimension of self, with the indices k given in dims (if provided).

        Parameters
        ----------
        matrices : numpy.ndarray or list of numpy.ndarray
            The diagonals of the matrices to use in the product.
        dims : list or numpy.ndarray, optional
            Indices of the contractions. The default is None, indicating all
            the dimensions.

        Returns
        -------
        FullTensor
            The tensor after the contractions with the matrices.

        '''
        if dims is None:
            assert isinstance(matrices, list), 'matrices should be a list.'
            assert len(matrices) == self.order, \
                'len(matrices) must be self.order.'
            dims = range(self.order)
        else:
            dims = np.array(dims)
            if not isinstance(matrices, list):
                matrices = [matrices]
            assert len(matrices) == dims.size, \
                'len(matrices) must be equal to dims.size.'

        matrices = [FullTensor(np.diag(np.reshape(x, [-1])))
                    for x in matrices]
        return self.tensor_matrix_product(matrices, dims)

    def matricize(self, dims1, dims2=None):
        '''
        Return the matricization of the tensor.

        Parameters
        ----------
        dims1 : list or numpy.ndarray
            The dimensions of the tensor corresponding to the first dimension
            of the matricization.
        dims2 : list or numpy.ndarray, optional
            The dimensions of the tensor corresponding to the first dimension
            of the matricization. The default is None, for which they are
            deduced from dims1.

        Returns
        -------
        FullTensor
            The matricization of the tensor.

        '''
        dims1 = np.atleast_1d(dims1)
        if dims1.size == 1 and dims1 == -1:
            dims1 = np.array([self.order-1])
        if dims2 is None:
            dims2 = tensap.fast_setdiff(np.arange(self.order), dims1)
        else:
            dims2 = np.atleast_1d(dims2)
        shape1 = [self.shape[i] for i in dims1]
        shape2 = [self.shape[i] for i in dims2]

        tensor = FullTensor(self)
        tensor = tensor.transpose(np.concatenate((dims1, dims2)))
        tensor = tensor.reshape([np.prod(shape1), np.prod(shape2)])
        return FullTensor(tensor)

    def outer_product_eval_diag(self, tensor2, dims1, dims2, diag=False):
        '''
        Compute the diagonal of the outer product of two tensors.

        Equivalent to
        self.tensordot_eval_diag(tensor2, None, None, dims1, dims2, diag)

        Parameters
        ----------
        tensor2 : FullTensor
            The second tensor of the product.
        dims1 : list or numpy.ndarray
            Indices of the first tensor for the evaluation of the diagonal.
        dims2 : list or numpy.ndarray,
            Indices of the second tensor for the evaluation of the diagonal.
        diag : bool, optional
            Boolean enabling the evaluation of multiple diagonals. The default
            is False.

        Returns
        -------
        FullTensor
            The evaluated tensor.

        '''
        return self.tensordot_eval_diag(tensor2, None, None,
                                        dims1, dims2, diag)

    def principal_components(self, parameter=None):
        '''
        Compute the principal components of an order-2 tensor.

        Parameters
        ----------
        parameter : float or int, optional
            A parameter controlling the number of principal components.
            - If it is an integer, the number of principal components is the
            minimum between parameter and self.shape[0].
            - If it is a float smaller than 1, the number of principal
            components is determined such that ||x - VV'x||_F < t ||x||_F,
            with x the tensor, V the matrix of principal components, t the
            parameter, V' the transpose of the matrix V and ||.||_F the
            Frobenius norm.
            The default is self.shape[0].

        Returns
        -------
        principal_components : numpy.ndarray
            The principal components of the tensor.
        singular_values : numpy.ndarray
            The diagonal matrix of the associated singular values.

        '''
        assert self.order == 2, 'The order of the tensor must be 2.'
        if parameter is None or parameter > self.shape[0]:
            parameter = self.shape[0]

        if parameter < 1:
            truncator = tensap.Truncator(tolerance=parameter, max_rank=np.inf)
        else:
            truncator = tensap.Truncator(tolerance=0, max_rank=parameter)
        tensor = truncator.truncate(self)
        principal_components = tensor.space[0]
        singular_values = np.diag(tensor.core.data)
        return principal_components, singular_values

    def alpha_principal_components(self, alpha, parameter=None):
        '''
        Compute the alpha-principal components of a tensor.

        Return the principal components of the alpha-matricization
        M_alpha(self) of the tensor self of order d.

        See also the method principal_components.

        Parameters
        ----------
        alpha : int
            The index of the alpha-matricization.
        parameter : float or int, optional
            A parameter controlling the number of principal components.
            The default is M_alpha(self).shape[0].

        Returns
        -------
        principal_components : numpy.ndarray
            The principal components of the tensor.
        singular_values : numpy.ndarray
            The diagonal matrix of the associated singular values.

        '''
        principal_components, singular_values = \
            self.matricize(alpha).principal_components(parameter)
        return principal_components, singular_values

    def singular_values(self):
        '''
        Compute the higher-order singular values of a tensor (the collection
        of singular values of d different matricizations).

        Returns
        -------
        sin_val : numpy.ndarray or list of numpy.ndarray.
            The higher-order singular values.

        '''
        if self.order == 2:
            sin_val = np.linalg.svd(self.data, compute_uv=False)
        else:
            sin_val = []
            for ind in range(self.order):
                mat = self.matricize(ind)
                sin_val.append(np.linalg.svd(mat.data,
                                             compute_uv=False))
        return sin_val

    @staticmethod
    def create(generator, shape):
        '''
        Create a FullTensor of shape shape using a given generator.

        Parameters
        ----------
        generator : function
            Function generating a numpy.ndarray, given a shape.
        shape : numpy.ndarray or list
            The shape of the tensor.

        Returns
        -------
        FullTensor
            The created tensor.

        '''
        return FullTensor(generator(np.atleast_1d(shape)))

    @staticmethod
    def zeros(shape):
        '''
        Create a FullTensor of shape shape with entries equal to 0.

        Parameters
        ----------
        shape : numpy.ndarray or list
            The shape of the tensor.

        Returns
        -------
        FullTensor
            The created tensor.

        '''
        return FullTensor.create(np.zeros, shape)

    @staticmethod
    def ones(shape):
        '''
        Create a FullTensor of shape shape with entries equal to 1.

        Parameters
        ----------
        shape : numpy.ndarray or list
            The shape of the tensor.

        Returns
        -------
        FullTensor
            The created tensor.

        '''
        return FullTensor.create(np.ones, shape)

    @staticmethod
    def randn(shape):
        '''
        Create a FullTensor of shape shape with i.i.d. entries drawn according
        to the standard gaussian distribution.

        Parameters
        ----------
        shape : numpy.ndarray or list
            The shape of the tensor.

        Returns
        -------
        FullTensor
            The created tensor.

        '''
        return FullTensor.create(lambda x: np.random.randn(*x), shape)

    @staticmethod
    def rand(shape):
        '''
        Create a FullTensor of shape shape with i.i.d. entries drawn according
        to the uniform distribution on [0, 1].

        Parameters
        ----------
        shape : numpy.ndarray or list
            The shape of the tensor.

        Returns
        -------
        FullTensor
            The created tensor.

        '''
        return FullTensor.create(lambda x: np.random.rand(*x), shape)

    @staticmethod
    def diag(diag, order):
        '''
        Create a diagonal tensor x of order order, such that
        x[i, ..., i] = diag[i] for i = 0, ..., diag.size - 1.

        Parameters
        ----------
        diag : list or numpy.ndarray
            The diagonal of the tensor.
        order : int
            The order of the tensor.

        Returns
        -------
        FullTensor
            The created tensor.

        '''
        diag = np.atleast_1d(diag)

        ones_v = np.ones(order, dtype=int)
        shape_v = diag.size * ones_v

        data = np.zeros(shape_v)
        for ind, diag_ind in enumerate(diag):
            data[tuple(ind * ones_v)] = diag_ind
        return FullTensor(data)



# Copyright (c) 2020, Anthony Nouy, Erwan Grelier
# This file is part of tensap (tensor approximation package).

# tensap is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# tensap is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with tensap.  If not, see <https://www.gnu.org/licenses/>.

'''
Module canonical_tensor.

'''
from copy import deepcopy
import numpy as np
import tensap


class CanonicalTensor:
    '''
    Class CanonicalTensor.

    Attributes
    ----------
    core : tensap.DiagonalTensor
        The core of the canonical tensor.
    space : numpy.ndarray
        The tensor space.
    order : int
        The order of the tensor.
    shape : numpy.ndarray
        The shape of the tensor.
    is_orth : bool
        Boolean indicating, if is_orth = True, the dimension mu for which
        the mu-matricization of the tensor is orthogonal.

    '''

    def __init__(self, space, core):
        '''
        Constructor for the class CanonicalTensor

        Parameters
        ----------
        space : numpy.ndarray or list
        The tensor space.
        core : tensap.DiagonalTensor or list or numpy.ndarray
        The core of the canonical tensor.

        Returns
        -------
        None.

        '''
        assert isinstance(space, (list, np.ndarray)), \
            'The input space must be a list or a numpy.ndarray.'

        assert isinstance(core, (list, np.ndarray, tensap.DiagonalTensor)), \
            ('The input core must be a list or a numpy.ndarray or a ' +
             'tensap.DiagonalTensor.')

        if not isinstance(core, tensap.DiagonalTensor):
            core = tensap.DiagonalTensor(core, len(space))

        self.core = core
        self.space = space
        self.order = core.order
        self.shape = np.array([x.shape[0] for x in self.space])
        self.is_orth = False

    def __add__(self, arg):
        core = np.concatenate((self.core.data, arg.core.data))
        space = [np.hstack((x, y)) for x, y in zip(list(self.space),
                                                   list(arg.space))]
        return CanonicalTensor(space, core)

    def __radd__(self, arg):
        return self + arg

    def __sub__(self, arg):
        return self + (-arg)

    def __rsub__(self, arg):
        return -self + arg

    def __neg__(self):
        return CanonicalTensor(list(self.space), -self.core.data)

    def __repr__(self):
        return ('<{} CanonicalTensor:{n}' +
                '{t}order = {},{n}' +
                '{t}shape = {},{n}' +
                '{t}is_orth = {}>').format('x'.join(map(str, self.shape)),
                                           self.order,
                                           self.shape,
                                           self.is_orth,
                                           t='\t', n='\n')

    @property
    def ndim(self):
        '''
        Compute the order of the tensor. Equivalent to self.order.

        Returns
        -------
        int
            The order of the tensor.

        '''
        return self.order

    def orth(self):
        '''
        Return an orthonormalized representation of the tensor.

        Returns
        -------
        out : CanonicalTensor
            The orthonormalized representation of the tensor.

        '''
        dims = range(self.order)
        qr = [np.linalg.qr(x) for x in self.space]
        space = [x[0] for x in qr]
        M = [x[1] for x in qr]
        core = self.core.tensor_matrix_product(M, dims)
        core = core.orth()[0]
        out = CanonicalTensor(space, self.core)
        out.core = core
        return out

    def full(self):
        '''
        Convert the object to a tensap.FullTensor.

        Returns
        -------
        tensap.FullTensor
            The canonical tensor as a tensap.FullTensor.

        '''
        return self.core.full().tensor_matrix_product(self.space)

    def numpy(self):
        '''
        Convert the CanonicalTensor to a numpy.ndarray.

        Returns
        -------
        numpy.ndarray
            The CanonicalTensor as a numpy.ndarray.

        '''
        return self.full().data

    def tensor_matrix_product(self, matrices, dims=None):
        '''
        Contract a tensor with matrices.

        The second dimension of the matrix matrices[k] is contracted with the
        k-th dimension of self, with the indices k given in dims (if provided).

        Parameters
        ----------
        matrices : numpy.ndarray or list of numpy.ndarray
            The matrices to use in the product.
        dims : list or numpy.ndarray, optional
            Indices of the contractions. The default is None, indicating all
            the dimensions.

        Returns
        -------
        CanonicalTensor
            The tensor after the contractions with the matrices.

        '''
        if dims is None:
            assert isinstance(matrices, (list, np.ndarray)), \
                'matrices should be a list or a numpy.ndarray.'
            assert len(matrices) == self.order, \
                'len(matrices) must be self.order.'
            dims = range(self.order)
        else:
            dims = np.atleast_1d(dims)
            if not isinstance(matrices, list):
                matrices = [matrices]
            assert len(matrices) == dims.size, \
                'len(matrices) must be equal to dims.size.'

        space = list(self.space)
        for i, dim in enumerate(dims):
            space[dim] = np.matmul(matrices[i], space[dim])
        return CanonicalTensor(space, np.array(self.core.data))

    def tensor_matrix_product_eval_diag(self, matrices, dims=None):
        '''
        Evaluate the diagonal of a tensor obtained by contraction with
        matrices.

        Provides the diagonal of the tensor obtained by contracting the tensor
        with matrices H[k] along dimensions dims(k)+1, for k = 0, ...,
        dims.size-1.

        Parameters
        ----------
        matrices : list
            The matrices to use in the product.
        dims : list or numpy.ndarray, optional
            Indices of the contractions. The default is None, indicating all
            the dimensions.

        Returns
        -------
        out : CanonicalTensor or tensap.FullTensor
            The diagonal of the contractions of the tensor with the matrices.

        '''
        return self.tensor_matrix_product(matrices, dims).eval_diag(dims)

    def eval_diag(self, dims=None):
        '''
        Extract the diagonal of the tensor.

        The tensor must be such that self.shape[mu] = n for all mu (in dims if
        provided).

        Parameters
        ----------
        dims : list of numpy.ndarray, optional
            The dimensions associated with the indices of the diagonal. The
            default is None,indicating that the indices refer to all the
            dimensions.

        Returns
        -------
        data : CanonicalTensor or tensap.FullTensor
            The evaluations of the diagonal of the tensor.

        '''
        if dims is None:
            is_none = True
            dims = np.arange(self.order)
        else:
            is_none = False
            dims = np.atleast_1d(dims)
            if dims.size == 1:
                print('Only one dimension: degenerate case for eval_diag, ' +
                      'returning the tensor itself.')
                return deepcopy(self)
            dims = np.sort(dims)
            new_dims = np.setdiff1d(range(self.order), dims[1:])

        out = self.space[dims[0]]
        for k in dims[1:]:
            out *= self.space[k]

        if is_none or dims.size == self.order:
            out = tensap.FullTensor(np.matmul(out, self.core.data))
        else:
            space = list(self.space)
            core = deepcopy(self.core)

            space[dims[0]] = out
            space = space[new_dims]
            core.shape = core.shape[new_dims]
            core.order = new_dims.size
            out = CanonicalTensor(space, core)
        return out

    def dot(self, tensor_2):
        '''
        Return the inner product of two tensors.

        Parameters
        ----------
        tensor2 : CanonicalTensor
            The second tensor of the inner products.

        Returns
        -------
        numpy.float
            The inner product of the two tensors.

        '''
        assert isinstance(tensor_2, tensap.CanonicalTensor), \
            'The second argument must be a tensap.CanonicalTensor.'
        matrices = [np.matmul(np.transpose(x), y) for
                    x, y in zip(self.space, tensor_2.space)]
        return self.core.dot_with_rank_one_metric(tensor_2.core, matrices)

    def norm(self, matrix=None):
        '''
        Compute the canonical norm of the CanonicalTensor.

        Returns
        -------
        numpy.float
            The norm of the tensor.

        '''
        if matrix is None:
            if self.is_orth:
                norm = np.linalg.norm(self.core.data)
            else:
                norm = np.sqrt(np.abs(self.dot(self)))
        else:
            raise NotImplementedError('Method not implemented.')
        return norm

    def storage(self):
        '''
        Return the storage complexity of the CanonicalTensor.

        Returns
        -------
        int
            The storage complexity of the CanonicalTensor.

        '''
        return self.core.data.size + np.sum([x.size for x in self.space])

    def sparse_storage(self):
        '''
        Return the sparse storage complexity of the CanonicalTensor.

        Returns
        -------
        int
            The sparse storage complexity of the CanonicalTensor.

        '''
        return np.count_nonzero(self.core.data) + \
            np.sum([np.count_nonzero(x) for x in self.space])

    def representation_rank(self):
        '''
        Return the representation rank of the tensor.

        Returns
        -------
        int
            The representation rank of the tensor.

        '''
        return self.core.data.size

    def parameter_gradient_eval_diag(self, mu, matrices=None):
        '''
        Compute the diagonal of the gradient of the tensor with respect to a
        given parameter.

        Parameters
        ----------
        mu : int
            Index of the parameter.
        matrices : list or numpy.array, optional
            Matrices with which to compute outer_product_eval_diag if alpha is
            associated with some dimensions. Useful for evaluation the gradient
            of a tensap.FunctionalTensor. The default is None, indicating
            identity matrices.

        Returns
        -------
        out : tensap.FullTensor
            The diagonal of the gradient of the tensor with respect to
            the parameter with index mu.

        '''
        rank = len(self.core.data)
        if mu == self.order + 1:
            N = self.space[0].shape[0]
            out = np.ones((N, rank))
            for nu in range(self.order):
                out *= self.space[nu]
            out = tensap.FullTensor(out)
        else:
            no_mu = np.setdiff1d(np.arange(1, self.order+1), mu)
            N = self.space[no_mu[0]-1].shape[0]
            f_mu = np.ones((N, rank))
            for nu in no_mu:
                f_mu *= self.space[nu-1]
            if matrices is not None:
                out = tensap.FullTensor(f_mu).outer_product_eval_diag(
                    tensap.FullTensor(matrices[mu-1]), 0, 0)
            else:
                out = tensap.FullTensor(f_mu).outer_product_eval_diag(
                    tensap.FullTensor(np.eye(self.shape[mu-1])), [], [], True)
        return out

    def tree_based_tensor(self, tree, is_active_node=None):
        '''
        Convert a CanonicalTensor to a tensap.TreeBasedTensor with given
        dimension tree and active nodes.

        Parameters
        ----------
        tree : tensap.DimensionTree
            The dimension tree.
        is_active_node : list or numpy.ndarray, optional
            Booleans indicating if the nodes are active. The default is None,
            settings all the nodes active.

        Returns
        -------
        out : tensap.TreeBasedTensor
            The canonical tensor converted into a tree-based tensor.

        '''

        rank = self.core.shape[0]
        tensors = np.empty(tree.nb_nodes, dtype=object)
        for nod in np.arange(1, tree.nb_nodes+1):
            ch = tree.children(nod)
            if tree.parent(nod) == 0:
                order = ch.size
                tensors[nod-1] = tensap.FullTensor.diag(self.core.data, order)
            else:
                order = ch.size + 1
                tensors[nod-1] = tensap.FullTensor.diag(np.ones(rank), order)

        tensors[tree.dim2ind-1] = self.space
        out = tensap.TreeBasedTensor(tensors, tree)
        if is_active_node is not None:
            out = out.inactivate_nodes(
                np.nonzero(np.logical_not(is_active_node))[0]+1)
        return out

    @staticmethod
    def create(generator, rank, shape):
        '''
        Create a FullTensor of rank rank and shape shape using a given
        generator.

        Parameters
        ----------
        generator : function
            Function generating a numpy.ndarray, given a shape.
        rank : int
            The rank of the tensor.
        shape : numpy.ndarray or list
            The shape of the tensor.

        Returns
        -------
        CanonicalTensor
            The created tensor.

        '''
        space = [generator([x, rank]) for x in shape]
        return CanonicalTensor(space, np.ones(rank))

    @staticmethod
    def zeros(rank, shape):
        '''
        Create a FullTensor of rank rank and shape shape with entries equal to
        0.

        Parameters
        ----------
        rank : int
            The rank of the tensor.
        shape : numpy.ndarray or list
            The shape of the tensor.

        Returns
        -------
        CanonicalTensor
            The created tensor.

        '''
        return CanonicalTensor.create(np.zeros, rank, shape)

    @staticmethod
    def ones(rank, shape):
        '''
        Create a FullTensor of rank rank and shape shape with entries equal to
        1.

        Parameters
        ----------
        rank : int
            The rank of the tensor.
        shape : numpy.ndarray or list
            The shape of the tensor.

        Returns
        -------
        CanonicalTensor
            The created tensor.

        '''
        return CanonicalTensor.create(np.ones, rank, shape)

    @staticmethod
    def rand(rank, shape):
        '''
        Create a FullTensor of rank rank and shape shape with i.i.d. entries
        drawn according to the uniform distribution on [0, 1].

        Parameters
        ----------
        rank : int
            The rank of the tensor.
        shape : numpy.ndarray or list
            The shape of the tensor.

        Returns
        -------
        CanonicalTensor
            The created tensor.

        '''
        return CanonicalTensor.create(lambda x: np.random.rand(*x), rank,
                                      shape)

    @staticmethod
    def randn(rank, shape):
        '''
        Create a FullTensor of rank rank and shape shape with i.i.d. entries
        drawn according to the standard gaussian distribution.

        Parameters
        ----------
        rank : int
            The rank of the tensor.
        shape : numpy.ndarray or list
            The shape of the tensor.

        Returns
        -------
        CanonicalTensor
            The created tensor.

        '''
        return CanonicalTensor.create(lambda x: np.random.randn(*x), rank,
                                      shape)



# Copyright (c) 2020, Anthony Nouy, Erwan Grelier
# This file is part of tensap (tensor approximation package).

# tensap is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# tensap is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with tensap.  If not, see <https://www.gnu.org/licenses/>.

'''
Module linear_model_learning_custom_loss.

'''

from numpy import arange, finfo
import numpy as np
import tensap
try:
    import tensorflow as tf
    cond = True
except ImportError:
    cond = False


class LinearModelLearningCustomLoss(tensap.LinearModelLearning):
    '''
    Class LinearModelLearningCustomLoss.

    Attributes
    ----------
    optimizer : tensorflow.keras.optimizers.Optimizer
          The optimizer used to solve the learning problem. The default is Adam.
    initial_guess : numpy.ndarray or tensorflow.Tensor
          The initial guess used as a starting point of the optimization
          algorithm. The default is a tensor with components drawn according
          to a standard normal random variable.
    options : dict
        Options for the optimizer:

        - max_iter: the maximum number of iterations of an iterative
          minimization algorithm,
        - stagnation: the value of a stopping criterion based on the
          relative stagnation between two iterates

    '''

    def __init__(self, custom_loss):
        '''
        Constructor for the class LinearModelLearningCustomLoss.

        Parameters
        ----------
        custom_loss : tap.CustomLossFunction
            The loss function.
        '''
        if not cond:
            raise ImportError('Package tensorflow must be installed to ' +
                              'use LinearModelLearningCustomLoss.')
        super().__init__(custom_loss)
        self.optimizer = tf.keras.optimizers.Adam()
        self.initial_guess = None

        self.options = {'max_iterations': 1e3,
                        'stagnation': finfo(float).eps}

    def solve(self):
        '''
        Solution of the minimization problem.

        Returns
        -------
        sol : numpy.ndarray or tensap.FunctionalBasisArray
            The solution of the minimization problem.
        output : dict
            Outputs of the algorithm.
        '''
        self.initialize()

        if self.initial_guess is None:
            self.initial_guess = tf.random.normal([self.basis_eval.shape[1]],
                                                  dtype=tf.float64)
        else:
            self.initial_guess = tf.convert_to_tensor(self.initial_guess,
                                                      dtype=tf.float64)

        basis_eval = self.basis_eval
        training_data = self.training_data

        def risk():
            fun_eval = tf.squeeze(tf.tensordot(basis_eval, var, [1, 0]))
            out = self.loss_function.risk_estimation(fun_eval,
                                                     training_data)
            return out

        var = tf.Variable(self.initial_guess, dtype=tf.float64)
        for it in arange(self.options['max_iterations']):
            var0 = var.numpy()
            self.optimizer.minimize(risk, var_list=[var])

            stagnation = (tf.linalg.norm(var0 - var) /
                          tf.linalg.norm(var0)).numpy()
            if stagnation < self.options['stagnation']:
                break

        sol = var.numpy()

        output = {}
        if self.test_error:
            f_eval = np.matmul(self.basis_eval_test, sol)
            test_error = self.loss_function.test_error(
                f_eval, self.test_data)
            if isinstance(test_error, tf.Tensor):
                test_error = test_error.numpy()
            output['test_error'] = test_error

        if self.basis is not None:
            if np.ndim(sol) == 1:
                sol = tensap.FunctionalBasisArray(sol, self.basis)
            else:
                sol = tensap.FunctionalBasisArray(sol, self.basis,
                                                  sol.shape[1])

        return sol, output



# Copyright (c) 2020, Anthony Nouy, Erwan Grelier
# This file is part of tensap (tensor approximation package).

# tensap is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# tensap is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with tensap.  If not, see <https://www.gnu.org/licenses/>.

'''
Tutorial on principal component analysis for low rank approximation of tensors.

See the following article:
Anthony Nouy. Higher-order principal component analysis for the approximation
of tensors in tree-based low-rank formats. Numerische Mathematik,
141(3):743--789, Mar 2019.

'''

import numpy as np
import tensap

# %% Creating a function providing the entries of a tensor
D = 5
N = 10
SIZE = np.full(D, N)


def fun(i):
    return np.cos(i[:, 0]/SIZE[0]) + 1/(1+(i[:, 1]/SIZE[1])**2 +
                                        (i[:, 2]/SIZE[2])**4) + i[:, 2]/SIZE[2]


X = tensap.random_multi_indices(SIZE)
TOL = 1e-8

# %% HOPCA (PCA for each dimension, provides reduced spaces)
print('--- Higher order PCA ---')
TPCA = tensap.TensorPrincipalComponentAnalysis()
TPCA.pca_sampling_factor = 1
TPCA.pca_adaptive_sampling = False

print('\nPrescribed tolerance')
TPCA.tol = TOL
SUB_BASES, OUTPUTS = TPCA.hopca(fun, SIZE)
print('Number of evaluations = \n%s' % [x['number_of_evaluations'] for x
                                        in OUTPUTS])
print('Ranks {1, ..., d} = \n%s' % [x.shape[1] for x in SUB_BASES])

print('\nPrescribed ranks')
TPCA.tol = np.inf
TPCA.max_rank = np.random.randint(1, 5, D)
SUB_BASES, OUTPUTS = TPCA.hopca(fun, SIZE)
print('Number of evaluations = \n%s' % [x['number_of_evaluations'] for x
                                        in OUTPUTS])
print('Ranks {1, ..., d} = \n%s' % [x.shape[1] for x in SUB_BASES])

# %% Approximation in Tucker Format
print('\n--- Approximation in Tucker format ---')
TPCA = tensap.TensorPrincipalComponentAnalysis()
TPCA.pca_sampling_factor = 1
TPCA.pca_adaptive_sampling = False

print('\nPrescribed ranks')
TPCA.tol = np.inf
TPCA.max_rank = np.random.randint(1, 5, D)
F, OUTPUT = TPCA.tucker_approximation(fun, SIZE)

print('Number of evaluations = %i' % OUTPUT['number_of_evaluations'])
print('Prescribed ranks = %s' % TPCA.max_rank)
print('Ranks = %s' % F.ranks[F.tree.dim2ind-1])
X_TEST = X.random(1e4)
F_X_TEST = F.eval_at_indices(X_TEST)
Y_TEST = fun(X_TEST)
print('Error = %2.5e' % (np.linalg.norm(Y_TEST - F_X_TEST) /
                         np.linalg.norm(Y_TEST)))

print('\nPrescribed tolerance')
TPCA.tol = TOL
TPCA.max_rank = np.inf
F, OUTPUT = TPCA.tucker_approximation(fun, SIZE)

print('Number of evaluations = %i' % OUTPUT['number_of_evaluations'])
print('Ranks = %s' % F.ranks[F.tree.dim2ind-1])
X_TEST = X.random(1e4)
F_X_TEST = F.eval_at_indices(X_TEST)
Y_TEST = fun(X_TEST)
print('Error = %2.5e' % (np.linalg.norm(Y_TEST - F_X_TEST) /
                         np.linalg.norm(Y_TEST)))

# %% Approximation in tree based format
print('\n--- Approximation in tree based format ---')
TPCA = tensap.TensorPrincipalComponentAnalysis()
TPCA.pca_sampling_factor = 1
TPCA.pca_adaptive_sampling = False
TREE = tensap.DimensionTree.balanced(D)

print('\nPrescribed ranks')
TPCA.tol = np.inf
TPCA.max_rank = np.random.randint(1, 9, TREE.nb_nodes)
TPCA.max_rank[TREE.root-1] = 1
F, OUTPUT = TPCA.tree_based_approximation(fun, SIZE, TREE)

print('Number of evaluations = %i' % OUTPUT['number_of_evaluations'])
print('Storage = %i' % F.storage())
print('Prescribed ranks = %s' % TPCA.max_rank)
print('Ranks = %s' % F.ranks)
X_TEST = X.random(1e3)
F_X_TEST = F.eval_at_indices(X_TEST)
Y_TEST = fun(X_TEST)
print('Error = %2.5e' % (np.linalg.norm(Y_TEST - F_X_TEST) /
                         np.linalg.norm(Y_TEST)))

print('\nPrescribed tolerance')
TPCA.tol = TOL
TPCA.max_rank = np.inf
F, OUTPUT = TPCA.tree_based_approximation(fun, SIZE, TREE)

print('Number of evaluations = %i' % OUTPUT['number_of_evaluations'])
print('Storage = %i' % F.storage())
print('Ranks = %s' % F.ranks)
X_TEST = X.random(1e3)
F_X_TEST = F.eval_at_indices(X_TEST)
Y_TEST = fun(X_TEST)
print('Error = %2.5e' % (np.linalg.norm(Y_TEST - F_X_TEST) /
                         np.linalg.norm(Y_TEST)))

print('\nPrescribed tolerance (adaptive sampling)')
TPCA.tol = TOL
TPCA.pca_adaptive_sampling = True
TPCA.pca_sampling_factor = 1.2
TPCA.max_rank = np.inf
F, OUTPUT = TPCA.tree_based_approximation(fun, SIZE, TREE)

print('Number of evaluations = %i' % OUTPUT['number_of_evaluations'])
print('Storage = %i' % F.storage())
print('Ranks = %s' % F.ranks)
X_TEST = X.random(1e3)
F_X_TEST = F.eval_at_indices(X_TEST)
Y_TEST = fun(X_TEST)
print('Error = %2.5e' % (np.linalg.norm(Y_TEST - F_X_TEST) /
                         np.linalg.norm(Y_TEST)))

# %% Approximation in Tensor Train format
print('\n--- Approximation in tensor train format ---')
TPCA = tensap.TensorPrincipalComponentAnalysis()
TPCA.pca_sampling_factor = 1
TPCA.pca_adaptive_sampling = False

print('\nPrescribed tolerance')
TPCA.tol = TOL
TPCA.max_rank = np.inf
F, OUTPUT = TPCA.tt_approximation(fun, SIZE)

print('Number of evaluations = %i' % OUTPUT['number_of_evaluations'])
print('Storage = %i' % F.storage())
TT_RANKS = np.flip(F.ranks[F.is_active_node])
print('TT-ranks = %s' % TT_RANKS[:-1])
X_TEST = X.random(1e3)
F_X_TEST = F.eval_at_indices(X_TEST)
Y_TEST = fun(X_TEST)
print('Error = %2.5e' % (np.linalg.norm(Y_TEST - F_X_TEST) /
                         np.linalg.norm(Y_TEST)))

print('\nPrescribed ranks')
TPCA.tol = np.inf
TPCA.max_rank = np.random.randint(1, 9, D-1)
F, OUTPUT = TPCA.tt_approximation(fun, SIZE)

print('Number of evaluations = %i' % OUTPUT['number_of_evaluations'])
print('Storage = %i' % F.storage())
print('Prescribed ranks = %s' % TPCA.max_rank)
TT_RANKS = np.flip(F.ranks[F.is_active_node])
print('TT-ranks = %s' % TT_RANKS[:-1])
X_TEST = X.random(1e3)
F_X_TEST = F.eval_at_indices(X_TEST)
Y_TEST = fun(X_TEST)
print('Error = %2.5e' % (np.linalg.norm(Y_TEST - F_X_TEST) /
                         np.linalg.norm(Y_TEST)))



# Copyright (c) 2020, Anthony Nouy, Erwan Grelier
# This file is part of tensap (tensor approximation package).

# tensap is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# tensap is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with tensap.  If not, see <https://www.gnu.org/licenses/>.

'''
Tutorial on Tensorizer, TensorizedFunction and FunctionalTensorPrincipalComponentAnalysis.

'''

import numpy as np
import tensap


# %% Identification of a multivariate function f(x1, ..., xd) on (0,1)^d with a function
# g(i_1^1,...,i_1^d, ..., i_L^1,...,i_L^d, y1, ...,yd)
# x1 and x2 are identified with (i_1^1, ,..., i_L^1, y1) and (i_1^d, ...., i_L^d, yd)
# through a Tensorizer

DIM = 4
L = 13  # Resolution
B = 2  # Scaling factor

T = tensap.Tensorizer(B, L, DIM)
T.ordering_type = 2
FUN = tensap.UserDefinedFunction('1/(1+x0+2*x1+3*x2+4*x3)', DIM)
FUN.evaluation_at_multiple_points = True
TENSORIZED_FUN = T.tensorize(FUN)

# Interpolation of the function in the tensor product feature space
DEGREE = 2
BASES = T.tensorized_function_functional_bases(DEGREE)



# %% Learning with PCA based algorithm
FPCA = tensap.FunctionalTensorPrincipalComponentAnalysis()
FPCA.pca_sampling_factor = 20
FPCA.bases = BASES
FPCA.display = False
TREE = tensap.DimensionTree.balanced(TENSORIZED_FUN.fun.dim)
FPCA.tol = 1e-8
FPCA.max_rank = np.inf
FUN_TB, OUTPUT = FPCA.tree_based_approximation(TENSORIZED_FUN, TREE)
TENSORIZED_FUN_TB = tensap.TensorizedFunction(FUN_TB, T)
X_TEST = T.X.random(1000)
F_X_TEST = TENSORIZED_FUN_TB(X_TEST)
Y_TEST = FUN(X_TEST)
ERR_L2 = np.linalg.norm(Y_TEST - F_X_TEST) / np.linalg.norm(Y_TEST)
print('Mean squared error = %2.5e' % ERR_L2)





# Copyright (c) 2020, Anthony Nouy, Erwan Grelier
# This file is part of tensap (tensor approximation package).

# tensap is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# tensap is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with tensap.  If not, see <https://www.gnu.org/licenses/>.

'''
Module copulas.

'''

from abc import ABC
import numpy as np


class Copula(ABC):
    pass


class IndependentCopula(Copula):
    def pdf(self, u):
        return np.ones((np.shape(u)[0], 1))

    def cdf(self, u):
        return np.prod(u, axis=1)



# Copyright (c) 2020, Anthony Nouy, Erwan Grelier
# This file is part of tensap (tensor approximation package).

# tensap is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# tensap is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with tensap.  If not, see <https://www.gnu.org/licenses/>.


'''
Module polynomial_functional_basis.

'''

import numpy as np
import tensap


class PolynomialFunctionalBasis(tensap.FunctionalBasis):
    '''
    Class PolynomialFunctionalBasis.

    Attributes
    ----------
    basis : tensap.UnivariatePolynomials
        The polynomials associated with the basis.

    indices : list or numpy.ndarray
        The indices of the selected polynomials.

    measure : tensap.Measure
        The measure associated with basis.

    is_orthonormal : bool
        Boolean equal to true, indicating that the basis is orthonormal with
        respect to measure.

    '''

    def __init__(self, basis, indices):
        '''
        Constructor for the class PolynomialFunctionalBasis.

        Parameters
        ----------
        basis : tensap.UnivariatePolynomials
            The polynomials associated with the basis.
        indices : list or numpy.ndarray
            The indices of the selected polynomials.

        Returns
        -------
        None.

        '''
        tensap.FunctionalBasis.__init__(self)
        self.basis = basis
        self.indices = np.array(indices)
        self.measure = basis.measure

        if isinstance(basis, (tensap.OrthonormalPolynomials,
                              tensap.ShiftedOrthonormalPolynomials)):
            self.is_orthonormal = True

    def __repr__(self):
        return ('<{}:{n}' +
                '{t}basis = {},{n}' +
                '{t}indices = {},{n}' +
                '{t}measure = {},{n}' +
                '{t}is_orthonormal = {}>').format(self.__class__.__name__,
                                                  self.basis,
                                                  self.indices,
                                                  self.measure,
                                                  self.is_orthonormal,
                                                  t='\t', n='\n')

    def __eq__(self, basis_2):
        if not isinstance(basis_2, PolynomialFunctionalBasis):
            out = False
        else:
            out = self.basis == basis_2.basis and \
                np.array_equal(self.indices, basis_2.indices)
        return out

    def cardinal(self):
        return self.indices.size

    def eval(self, x):
        '''
        Evaluate the polynomials of self.basis of degrees in self.indices at
        points x.

        Parameters
        ----------
        x : list or numpy.ndarray
            The points at which the basis functions are to be evaluated.

        Returns
        -------
        numpy.ndarray
            The evaluations of the polynomials of self.basis of degrees in
            self.indices at points x.

        '''
        return self.basis.polyval(self.indices, x)

    def kron(self, q):
        m = np.max(self.indices) + np.max(q.indices)
        b = PolynomialFunctionalBasis(self.basis, range(m))
        b.measure = self.measure

        def fun(x):
            return np.reshape(np.tile(np.expand_dims(self.eval(x), 2),
                                      (1, 1, q.cardinal())) *
                              np.tile(np.transpose(
                                  np.expand_dims(q.eval(x), 2), [0, 2, 1]),
                                  (1, self.cardinal(), 1)),
                              (np.shape(x)[0], self.cardinal()*q.cardinal()),
                              order='F')
        measure = b.measure
        int_rule = measure.gauss_integration_rule(2*m)
        return tensap.SubFunctionalBasis(b,
                                         self.projection(fun, int_rule).data)

    def derivative(self, k, measure=None):
        '''
        Compute the k-th order derivative of the functions of the basis
        projected on itself.

        Parameters
        ----------
        k : int
            The order of the derivative.
        measure : tensap.Measure, optional
            The measure used fot the projection. The default is None,
            indicating to use self.measure if it is a tensap.RandomVariable.

        Raises
        ------
        ValueError
            If no Measure is provided or can be extracted from self.

        Returns
        -------
        tensap.SubFunctionalBasis
            The k-th order derivative of the functions of the basis projected
            on itself.

        '''
        if measure is None:
            assert self.measure is not None, 'Must specify a Measure.'
            measure = self.measure

        nb_pts = int(np.ceil(np.max(self.indices) - (k-1)/2))
        int_rule = measure.gauss_integration_rule(nb_pts)
        def fun(x): return self.basis.dn_polyval(k, self.indices, x)
        return self.projection(fun, int_rule).sub_functional_basis()

    def gradient(self, measure=None):
        '''
        Compute the first order derivative of the functions of the basis
        projected on itself.

        Parameters
        ----------
        measure : tensap.Measure, optional
            The measure used fot the projection. The default is None,
            indicating to use self.measure if it is a tensap.RandomVariable.

        Returns
        -------
        tensap.SubFunctionalBasis
            The first order derivative of the functions of the basis projected
             on itself.

        '''
        return self.derivative(1, measure)

    def hessian(self, measure):
        '''
        Compute the second order derivative of the functions of the basis
        projected on itself.

        Parameters
        ----------
        measure : tensap.Measure, optional
            The measure used fot the projection. The default is None,
            indicating to use self.measure if it is a tensap.RandomVariable.

        Returns
        -------
        tensap.SubFunctionalBasis
            The second order derivative of the functions of the basis projected
             on itself.

        '''
        return self.derivative(2, measure)

    def eval_derivative(self, k, x):
        '''
        Evaluate the k-th order derivative of the functions of the basis at the
        points x.

        Parameters
        ----------
        k : int
            The order of the derivative.
        x : list or numpy.ndarray
            The points at which the k-th derivative of the basis functions are
            to be evaluated.

        Returns
        -------
        numpy.ndarray
            The evaluations of the k-th derivative of the polynomials of
            self.basis of degrees in self.indices at points x.

        '''
        return self.basis.dn_polyval(k, self.indices, x)

    def gram_matrix(self, measure=None):
        '''
        Compute the Gram matrix of the basis. The Gram matrix is the matrix of
        the dot products between each possible couple of basis functions. The
        dot product in the dimension i is computed according to measure if
        provided, or according to self.measure otherwise.

        Parameters
        ----------
        measure : tensap.Measure, optional
            The measure according to which the dot product is computed. The
            default is None, indicating to use self.measure.

        Raises
        ------
        NotImplementedError
            If the attribute basis of self does not have a method moment.

        Returns
        -------
        numpy.ndarray
            The Gram matrix of the basis.

        '''
        if measure is None:
            assert self.measure is not None, 'Must specify a Measure.'
            measure = self.measure
        try:
            ind = np.reshape(self.indices, [-1, 1])
            rep = np.hstack((
                np.reshape(np.tile(ind, [len(ind), 1]), [-1, 1]),
                np.reshape(np.tile(ind, [1, len(ind)]), [-1, 1])))
            return np.reshape(self.basis.moment(rep, measure), [len(ind)]*2)
        except AttributeError:
            raise NotImplementedError('Method not implemented.')

    def one(self):
        '''
        Return the coefficients associated with the basis so that it returns
        one.

        Returns
        -------
        list
            The list of coefficients so that the basis returns one.

        '''
        c, ind = self.basis.one()
        ok = np.isin(ind, self.indices)
        if not np.all(ok):
            raise ValueError('The constant 1 cannot be represented.')
        out = np.zeros(self.indices.size)
        out[np.isin(self.indices, ind)] = c
        return out

    def mean(self, measure=None):
        '''
        Return the expectation of the basis functions, accorging to measure if
        provided, and to self.measure ortherwise.

        Parameters
        ----------
        measure : None or tensap.RandomVariable, optional
            The measure according to which the mean is computed. The default is
            None, indicating to use self.measure.

        Returns
        -------
        numpy.ndarray
            The mean of the basis functions.

        '''
        return self.basis.mean(self.indices, measure)

    def random_variable(self):
        '''
        Return the tensap.RandomVariable associated with self if it exists.

        Returns
        -------
        out : tensap.RandomVariable or []
            The tensap.RandomVariable associated with self if it exists, and []
            otherwise.

        '''
        if hasattr(self.basis, 'random_variable'):
            out = self.basis.random_variable
        else:
            print('Empty random variable.')
            out = []
        return out

    def optimal_sampling_measure(self):
        # TODO optimal_sampling_measure
        raise NotImplementedError('Method not implemented.')

    def christoffel(self):
        # TODO christoffel
        raise NotImplementedError('Method not implemented.')

    def domain(self):
        return self.basis.domain()

    def ndim(self):
        return self.basis.ndim()



# Copyright (c) 2020, Anthony Nouy, Erwan Grelier
# This file is part of tensap (tensor approximation package).

# tensap is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# tensap is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with tensap.  If not, see <https://www.gnu.org/licenses/>.

'''
Module model_selection.

'''

import numpy as np
import tensap


class ModelSelection:
    '''
    Class ModelSelection.

    Attributes
    ----------
    pen_shape : function
        Function specifying the penalization shape.
    data : dict
        Dictionnary containing the data used for the model selection.
        data['complexity'] contains the complexities of the models, and
        data['empirical_risk'] the associated empirical risk values.
    gap_factor : int or float
        Multiplicative factor used in the slope heuristic.

    '''

    def __init__(self):
        '''
        Constructor for the class ModelSelection.

        Returns
        -------
        None.

        '''
        self.data = {'complexity': [],
                     'empirical_risk': []}
        self.pen_shape = lambda x: x
        self.gap_factor = 2

    def m_lambda(self, lbda):
        '''
        Compute the argument of the minimum of the penalized risk for given
        values of the penalization factor lbda.

        Parameters
        ----------
        lbda : list or numpy.ndarray
            The values of the penalization factor.

        Returns
        -------
        numpy.ndarray
            The argument of the minimum of the penalized risk for the values of
            the penalization factor.

        '''
        comp = np.array(self.data['complexity'])
        risk = np.array(self.data['empirical_risk'])
        return np.array([np.argmin(risk + l * self.pen_shape(comp)) for
                         l in np.atleast_1d(lbda)])

    def lambda_path(self):
        '''
        Return the path of possible values of lambda and associated arguments
        of the minimum of the penalized risk.


        Returns
        -------
        numpy.ndarray
            The path of possible values of lambda.
        numpy.ndarray
            The path of the arguments of the minimum of the penalized risk
            associated with the path of possible values of lambda.

        '''
        comp = np.array(self.data['complexity'])
        risk = np.array(self.data['empirical_risk'])

        path = []
        m_path = []
        lambda_0 = 0
        lambda_current = 0
        ind = np.argmin(risk)
        ok = True

        while ok:
            with np.errstate(all='ignore'):
                lambda_current = (risk - risk[ind]) / \
                    (self.pen_shape(comp[ind]) - self.pen_shape(comp))
                lambda_current[lambda_current <= lambda_0] = np.nan
                lambda_current[lambda_current == np.inf] = np.nan

            ok = not np.all(np.isnan(lambda_current))
            if ok:
                lambda_0 = np.nanmin(lambda_current)
                if not np.isnan(lambda_0):
                    ind = np.nanargmin(lambda_current)
                    path.append(lambda_0)
                    m_path.append(ind)

        if path == []:
            path = 0
            m_path = 0
        else:
            path = np.atleast_1d(path)
            m_path = np.atleast_1d(m_path)
            ind_min = np.argmin(risk)
            path = np.concatenate(([path[0]/2], path, [path[-1]*2]))
            m_path = np.concatenate(([ind_min], m_path, [m_path[-1]]))
        return np.array(path), np.array(m_path)

    def slope_heuristic(self, lambda_path=None, m_path=None):
        '''
        Apply the slope heuristic to the path of possible values of lambda to
        compute its optimal value lambda_hat and associated argument of the
        minimum of the penalized risk m_hat.

        Parameters
        ----------
        lambda_path : list or numpy.ndarray, optional
            The path of possible values of lambda. The default is None,
            indicating to compute it using the method lambda_path.
        m_path : list or numpy.ndarray, optional
            The path of the arguments of the minimum of the penalized risk
            associated with the path of possible values of lambda. The default
            is None, indicating to compute it using the method lambda_path.

        Returns
        -------
        lambda_hat : float
            The value of lambda determined using the slope heuristic.
        m_hat : int
            The model number associated with the value of lambda determined
            using the slope heuristic.
        lambda_path : numpy.ndarray
            The path of possible values of lambda.
        m_path : numpy.ndarray
            The path of the arguments of the minimum of the penalized risk
            associated with the path of possible values of lambda.

        '''
        comp = np.array(self.data['complexity'])

        # If all the complexities are equal, choose the first model
        if np.all(np.diff(comp) == 0):
            lambda_hat = 0
            m_hat = 0
            lambda_path = lambda_hat
            m_path = m_hat
            return lambda_hat, m_hat, lambda_path, m_path

        if lambda_path is None and m_path is None:
            lambda_path, m_path = self.lambda_path()
        l_mid = (lambda_path[:-1] + lambda_path[1:]) / 2
        gaps = comp[self.m_lambda(l_mid[:-1])] - comp[self.m_lambda(l_mid[1:])]
        ind = np.argmax(gaps)
        lambda_hat = self.gap_factor * lambda_path[ind+1]
        m_hat = self.m_lambda(lambda_hat)
        return lambda_hat, m_hat, lambda_path, m_path

    @staticmethod
    def complexity(x, *args, **kwargs):
        '''
        Return the complexity associated to the input argument's type.

        See also complexity_functional_basis_array,
        complexity_tree_based_tensor.

        Parameters
        ----------
        x : list or numpy.ndarray or tensap.FunctionalBasisArray or
        tensap.FunctionalTensor or tensap.TreeBasedTensor
            The object(s) of which the complexity is computed.
        *args, **kwargs : tuples
            Additional parameters for the methods
            complexity_functional_basis_array and complexity_tree_based_tensor.

        Raises
        ------
        ValueError
            If the argument x is not of correct type.

        Returns
        -------
        float or list
            The complexity(ies) associated with the object(s).

        '''
        if isinstance(x, (list, np.ndarray)):
            return [ModelSelection.complexity(y, *args, **kwargs) for y in x]
        if isinstance(x, tensap.FunctionalBasisArray):
            return np.size(x.data)
        if isinstance(x, tensap.FunctionalTensor):
            return ModelSelection.complexity(x.tensor, *args, **kwargs)
        if isinstance(x, tensap.TreeBasedTensor):
            return ModelSelection.complexity_tree_based_tensor(x, *args,
                                                               **kwargs)
        raise ValueError('Wrong argument.')

    @staticmethod
    def complexity_functional_basis_array(x, fun=None):
        '''
        Return the complexity associated with a FunctionalBasisArray.

        Parameters
        ----------
        x : tensap.FunctionalBasisArray
            The FunctionalBasisArray of which the complexity is computed.
        fun : str, optional
            Name of the function applied to the array to extract the storage
            complexity. The default is 'storage'.

        Returns
        -------
        float
            The complexity associated with the FunctionalBasisArray.

        '''
        if fun is None:
            fun = 'storage'
        return eval('x.' + fun + '()')

    @staticmethod
    def complexity_tree_based_tensor(x, fun=None, c_type='standard'):
        '''
        Return the complexity associated with the TreeBasedTensor.

        Parameters
        ----------
        x : tensap.TreeBasedTensor
            The TreeBasedTensor of which the complexity is computed.
        fun : str, optional
            Name of the function applied to the array to extract the storage
            complexity. The default is 'storage'. Can also be 'sparse_storage'
            or 'sparse_leaves_storage' for instance.
        c_type : str, optional
            The complexity type. The default is 'standard'. Can also be
            'stiefel' or 'grassman'.

        Raises
        ------
        ValueError
            If the complexity type is neither 'standard' nor 'stiefel' nor
            'grassman'.

        Returns
        -------
        float or list
            The complexity(ies) associated with the TreeBasedTensor(s).

        '''
        if fun is None:
            fun = 'storage'

        if isinstance(x, (list, np.ndarray)):
            return ModelSelection.complexity(x, fun, c_type)
        if isinstance(x, tensap.FunctionalTensor):
            return ModelSelection.complexity(x.tensor, fun, c_type)
        assert isinstance(x, tensap.TreeBasedTensor), \
            'The first argument should be a TreeBasedTensor.'

        if c_type == 'standard':
            comp = eval('x.' + fun + '()')
        elif c_type == 'grassman':
            comp = eval('x.' + fun + '()') - np.sum(x.ranks**2)
        elif c_type == 'stiefel':
            comp = eval('x.' + fun + '()') - np.sum(x.ranks*(x.ranks+1)/2)
        else:
            raise ValueError('Wrong argument.')
        return comp



# Copyright (c) 2020, Anthony Nouy, Erwan Grelier
# This file is part of tensap (tensor approximation package).

# tensap is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# tensap is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with tensap.  If not, see <https://www.gnu.org/licenses/>.

'''
Module product_measure.

'''

import numpy as np
import tensap


class ProductMeasure(tensap.Measure):
    '''
    Class ProductMeasure.

    Attributes
    ----------
    measures : list or tensap.RandomVector
        List of Measure objects.

    '''

    def __new__(cls, measures):
        if np.all([isinstance(x, tensap.ProbabilityMeasure) for
                   x in measures]):
            return tensap.RandomVector(measures)
        else:
            return object.__new__(cls)

    def __init__(self, measures):
        '''
        Constructor for the class ProductMeasure.

        Parameters
        ----------
        measures : list or tensap.RandomVector
            List of Measure objects.

        Raises
        ------
        ValueError
            If the provided measures are not of type Measure.

        Returns
        -------
        None.

        '''
        if isinstance(measures, tensap.RandomVector):
            if not isinstance(measures.copula, tensap.IndependentCopula):
                print('The given Copula is replaced by an IndependentCopula.')
            measures = measures.random_variables
        elif not isinstance(measures, list):
            raise ValueError('measures must be a list of Measure.')
        self.measures = measures

    def __repr__(self):
        return self.__class__.__name__

    def __eq__(self, measure_2):
        return np.all([x == y for x, y in zip(self.measures,
                                              measure_2.measures)])

    def random_vector(self):
        '''
        Return, if self is a ProbabilityMeasure, the associated RandomVector.

        Returns
        -------
        tensap.RandomVector
            The RandomVector associated with self.

        '''
        assert np.all([isinstance(x, tensap.ProbabilityMeasure) for x in
                       self.measures]), \
            'The measures should be ProbabilityMeasure.'

        return tensap.RandomVector(self.measures)

    def mass(self):
        return np.prod([x.mass() for x in self.measures])

    def ndim(self):
        return np.sum([x.ndim() for x in self.measures])

    def support(self):
        return [x.support() for x in self.measures]

    def truncated_support(self):
        '''
        Return the truncated support of the measures of the ProductMeasure.

        Returns
        -------
        list
            The truncated support of the measures of the ProductMeasure.

        '''
        return [x.truncated_support() for x in self.measures]

    def marginal(self, ind):
        return ProductMeasure([self.measures[x] for x in ind])

    def pdf(self, x):
        # TODO pdf
        raise NotImplementedError('Method not implemented.')

    def random(self, x):
        # TODO random
        raise NotImplementedError('Method not implemented.')

    def random_sequential(self, x):
        # TODO random_sequential
        raise NotImplementedError('Method not implemented.')

    @staticmethod
    def duplicate(measure, dim):
        '''
        Create a ProductMeasure by duplicating dim times the provided measure.

        Parameters
        ----------
        measure : tensap.Measure
            The measure to be duplicated.
        dim : int
            The number of times the provided measure is duplicated.

        Returns
        -------
        tensap.ProductMeasure
            The created ProductMeasure.

        '''
        return ProductMeasure([measure]*dim)



# Copyright (c) 2020, Anthony Nouy, Erwan Grelier
# This file is part of tensap (tensor approximation package).

# tensap is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# tensap is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with tensap.  If not, see <https://www.gnu.org/licenses/>.

'''
Tutorial on learning in canonical tensor format.

'''

import tensap

# %% Function to approximate
CHOICE = 2
if CHOICE == 1:
    ORDER = 5
    X = tensap.RandomVector(tensap.NormalRandomVariable(), ORDER)

    def fun(x):
        return 1 / (10 + x[:, 0] + 0.5*x[:, 1])**2
elif CHOICE == 2:
    fun, X = tensap.multivariate_functions_benchmark('borehole')
    ORDER = X.size

# %% Approximation basis
DEGREE = 8
BASES = [tensap.PolynomialFunctionalBasis(X_TRAIN.orthonormal_polynomials(),
                                       range(DEGREE+1)) for
         X_TRAIN in X.random_variables]
BASES = tensap.FunctionalBases(BASES)

# %% Training and test samples
NUM_TRAIN = 1000
X_TRAIN = X.random(NUM_TRAIN)
Y_TRAIN = fun(X_TRAIN)

NUM_TEST = 10000
X_TEST = X.random(NUM_TEST)
Y_TEST = fun(X_TEST)

# %% Learning in canonical tensor format
SOLVER = tensap.CanonicalTensorLearning(ORDER, tensap.SquareLossFunction())
SOLVER.rank_adaptation = True
SOLVER.initialization_type = 'mean'
SOLVER.tolerance['on_error'] = 1e-6
SOLVER.alternating_minimization_parameters['stagnation'] = 1e-8
SOLVER.alternating_minimization_parameters['max_iterations'] = 100
SOLVER.linear_model_learning.regularization = False
SOLVER.linear_model_learning.basis_adaptation = True
SOLVER.bases = BASES
SOLVER.training_data = [X_TRAIN, Y_TRAIN]
SOLVER.display = True
SOLVER.alternating_minimization_parameters['display'] = False
SOLVER.test_error = True
SOLVER.test_data = [X_TEST, Y_TEST]
SOLVER.alternating_minimization_parameters['one_by_one_factor'] = False
SOLVER.alternating_minimization_parameters['inner_loops'] = 2
SOLVER.alternating_minimization_parameters['random'] = False
SOLVER.rank_adaptation_options['max_iterations'] = 20
SOLVER.model_selection = True
SOLVER.model_selection_options['type'] = 'test_error'

F, OUTPUT = SOLVER.solve()

TEST_ERROR = SOLVER.loss_function.test_error(F, [X_TEST, Y_TEST])
print('\nCanonical rank = %i, test error = %2.5e' % (len(F.tensor.core.data),
                                                   TEST_ERROR))



# Copyright (c) 2020, Anthony Nouy, Erwan Grelier
# This file is part of tensap (tensor approximation package).

# tensap is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# tensap is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with tensap.  If not, see <https://www.gnu.org/licenses/>.

'''
Tutorial on functional PCA for low rank approximation of multivariate
functions.

'''

import numpy as np
import tensap

# %% Choice of the function to approximate
CHOICE = 2
if CHOICE == 1:
    print('Henon-Heiles')
    D = 5
    FUN, X = tensap.multivariate_functions_benchmark('henon_heiles', D)

    DEGREE = 4
    BASES = [tensap.PolynomialFunctionalBasis(x.orthonormal_polynomials(),
                                              range(DEGREE+1)) for
             x in X.random_variables]

elif CHOICE == 2:
    print('Anisotropic function')
    D = 6
    X = tensap.RandomVector(tensap.UniformRandomVariable(-1, 1), D)

    def FUN(x):
        return 1/(10 + 2*x[:, 0] + x[:, 2] + 2*x[:, 3] - x[:, 4])**2

    DEGREE = 13
    BASES = [tensap.PolynomialFunctionalBasis(x.orthonormal_polynomials(),
                                              range(DEGREE+1)) for
             x in X.random_variables]

elif CHOICE == 3:
    print('Sinus of a sum')
    D = 10
    X = tensap.RandomVector(tensap.UniformRandomVariable(-1, 1), D)

    def FUN(x):
        return np.sin(np.sum(x[:, :10], 1))

    DEGREE = 17
    BASES = [tensap.PolynomialFunctionalBasis(x.orthonormal_polynomials(),
                                              range(DEGREE+1)) for
             x in X.random_variables]

elif CHOICE == 4:
    print('Composition of functions')
    D = 6
    X = tensap.RandomVector(tensap.UniformRandomVariable(-1, 1), D)

    TREE = tensap.DimensionTree.balanced(D)

    def FUN(x1, x2):
        return 1 / (1 + x1**2 + x2**2)
    FUN = tensap.CompositionalModelFunction(TREE, FUN, X)

    DEGREE = 4
    BASES = [tensap.PolynomialFunctionalBasis(x.orthonormal_polynomials(),
                                              range(DEGREE+1)) for
             x in X.random_variables]

elif CHOICE == 5:
    print('Borehole function')
    FUN, X = tensap.multivariate_functions_benchmark('borehole')
    D = X.size

    DEGREE = 14
    BASES = [tensap.PolynomialFunctionalBasis(x.orthonormal_polynomials(),
                                              range(DEGREE+1)) for
             x in X.random_variables]

elif CHOICE == 6:
    print('Tensorized function')
    R = 11  # Resolution
    B = 2  # Scaling factor
    D = R+1

    X = tensap.UniformRandomVariable(0, 1)
    Y = tensap.UniformRandomVariable(0, 1)

    def IFUN(x):
        return 1/(1+x)

    T = tensap.Tensorizer(B, R, 1, X, Y)
    FUN = T.tensorize(IFUN)
    FUN.fun.evaluation_at_multiple_points = True
    DEGREE = 1
    H = tensap.PolynomialFunctionalBasis(Y.orthonormal_polynomials(),
                                         range(DEGREE+1))
    BASES = T.tensorized_function_functional_bases(H)
    X = tensap.RandomVector(BASES.measure)

else:
    raise ValueError('Bad function choice.')

# %% HOPCA (PCA for each dimension, provides reduced spaces)
# PCA for each dimension to get principal subspaces
print('--- Higher order PCA ---')
FPCA = tensap.FunctionalTensorPrincipalComponentAnalysis()
FPCA.bases = BASES
FPCA.pca_sampling_factor = 1
FPCA.tol = 1e-10
SUB_BASES, OUTPUTS = FPCA.hopca(FUN)
print('Number of evaluations = \n%s' % [x['number_of_evaluations'] for x
                                        in OUTPUTS])
print('Ranks {1, ..., d} = \n%s' % [x.cardinal() for x in SUB_BASES])

# %% Approximation in Tucker Format
# PCA for each dimension to get principal subspaces and interpolation on the
# tensor product of principal subspaces
print('\n--- Approximation in Tucker format ---')
FPCA = tensap.FunctionalTensorPrincipalComponentAnalysis()
FPCA.pca_sampling_factor = 1
FPCA.pca_adaptive_sampling = True
FPCA.bases = BASES

print('\nPrescribed ranks')
FPCA.tol = np.inf
FPCA.max_rank = np.random.randint(1, 5, D)
F, OUTPUT = FPCA.tucker_approximation(FUN)

print('Number of evaluations = %i' % OUTPUT['number_of_evaluations'])
print('Storage = %i' % F.storage())
print('Prescribed ranks = %s' % FPCA.max_rank)
print('Ranks = %s' % F.tensor.ranks[F.tensor.tree.dim2ind-1])
X_TEST = X.random(1e4)
F_X_TEST = F(X_TEST)
Y_TEST = FUN(X_TEST)
print('Error = %2.5e' % (np.linalg.norm(Y_TEST - F_X_TEST) /
                         np.linalg.norm(Y_TEST)))

print('\nPrescribed tolerance')
FPCA.tol = 1e-10
FPCA.max_rank = np.inf
F, OUTPUT = FPCA.tucker_approximation(FUN)

print('Number of evaluations = %i' % OUTPUT['number_of_evaluations'])
print('Ranks = %s' % F.tensor.ranks[F.tensor.tree.dim2ind-1])
X_TEST = X.random(1e4)
F_X_TEST = F(X_TEST)
Y_TEST = FUN(X_TEST)
print('Error = %2.5e' % (np.linalg.norm(Y_TEST - F_X_TEST) /
                         np.linalg.norm(Y_TEST)))

# %% Approximation in tree based format
print('\n--- Approximation in tree based format ---')
FPCA = tensap.FunctionalTensorPrincipalComponentAnalysis()
FPCA.pca_sampling_factor = 20
FPCA.bases = BASES
TREE = tensap.DimensionTree.balanced(D)

print('\nPrescribed ranks')
FPCA.tol = np.inf
FPCA.max_rank = np.random.randint(1, 9, TREE.nb_nodes)
FPCA.max_rank[TREE.root-1] = 1
F, OUTPUT = FPCA.tree_based_approximation(FUN, TREE)

print('Number of evaluations = %i' % OUTPUT['number_of_evaluations'])
print('Storage = %i' % F.storage())
print('Prescribed ranks = %s' % FPCA.max_rank)
print('Ranks = %s' % F.tensor.ranks)
X_TEST = X.random(1e3)
F_X_TEST = F(X_TEST)
Y_TEST = FUN(X_TEST)
print('Error = %2.5e' % (np.linalg.norm(Y_TEST - F_X_TEST) /
                         np.linalg.norm(Y_TEST)))

print('\nPrescribed tolerance')
FPCA.tol = 1e-10
FPCA.max_rank = np.inf
F, OUTPUT = FPCA.tree_based_approximation(FUN, TREE)

print('Number of evaluations = %i' % OUTPUT['number_of_evaluations'])
print('Storage = %i' % F.storage())
print('Ranks = %s' % F.tensor.ranks)
X_TEST = X.random(1e3)
F_X_TEST = F(X_TEST)
Y_TEST = FUN(X_TEST)
print('Error = %2.5e' % (np.linalg.norm(Y_TEST - F_X_TEST) /
                         np.linalg.norm(Y_TEST)))

# %% Approximation in Tensor Train format
print('\n--- Approximation in tensor train format ---')
FPCA = tensap.FunctionalTensorPrincipalComponentAnalysis()
FPCA.projection_type = 'interpolation'
FPCA.pca_sampling_factor = 3
FPCA.pca_adaptive_sampling = True
FPCA.bases = BASES

print('\nPrescribed ranks')
FPCA.tol = np.inf
FPCA.max_rank = np.random.randint(1, 9, D-1)
F, OUTPUT = FPCA.tt_approximation(FUN)

print('Number of evaluations = %i' % OUTPUT['number_of_evaluations'])
print('Storage = %i' % F.storage())
print('Prescribed ranks = %s' % FPCA.max_rank)
TT_RANKS = np.flip(F.tensor.ranks[F.tensor.is_active_node])
print('TT-ranks = %s' % TT_RANKS[:-1])
X_TEST = X.random(1e3)
F_X_TEST = F(X_TEST)
Y_TEST = FUN(X_TEST)
print('Error = %2.5e' % (np.linalg.norm(Y_TEST - F_X_TEST) /
                         np.linalg.norm(Y_TEST)))

print('\nPrescribed tolerance')
FPCA.tol = 1e-4
FPCA.max_rank = np.inf
F, OUTPUT = FPCA.tt_approximation(FUN)

print('Number of evaluations = %i' % OUTPUT['number_of_evaluations'])
print('Storage = %i' % F.storage())
TT_RANKS = np.flip(F.tensor.ranks[F.tensor.is_active_node])
print('TT-ranks = %s' % TT_RANKS[:-1])
X_TEST = X.random(1e3)
F_X_TEST = F(X_TEST)
Y_TEST = FUN(X_TEST)
print('Error = %2.5e' % (np.linalg.norm(Y_TEST - F_X_TEST) /
                         np.linalg.norm(Y_TEST)))



# Copyright (c) 2020, Anthony Nouy, Erwan Grelier
# This file is part of tensap (tensor approximation package).

# tensap is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# tensap is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with tensap.  If not, see <https://www.gnu.org/licenses/>.

'''
Module functional_tensor_principal_component_analysis.

'''

from copy import deepcopy
import numpy as np
import tensap


class FunctionalTensorPrincipalComponentAnalysis:
    '''
    Class FunctionalTensorPrincipalComponentAnalysis: principal component
    analysis of multivariate functions based on
    TensorPrincipalComponentAnalysis for algebraic tensors.

    Attributes
    ----------
    tol : int or float or list or numpy.ndarray
        An array containing the prescribed relative precisions (the length
        depends on the format).
        If len(tol)==1, use the same value for all alpha.
        Set tol = inf to prescribe the rank.
    max_rank : int or list or numpy.ndarray
        An array containing the maximum alpha-ranks (the length depends on the
        format).
        If len(max_rank)==1, use the same value for all alpha.
        Set max_rank = np.inf to prescribe the precision.
    bases : tensap.FunctionalBases
        The functional bases used for the projection of the function.
    grid : tensap.FullTensorGrid
         FullTensorGrid for the projection of the function on the functional
         bases.
    display : bool
        Boolean specifying the verbosity of the methods.
    pca_sampling_factor : int
        A factor to determine the number of samples N for the estimation of
        the principal components (1 by default):

        - if prescribed precision, N = pca_sampling_factor*N_alpha,
        - if prescribed rank, N = pca_sampling_factor*t.

    pca_adaptive_sampling : bool
        Adaptive sampling to determine the principal components with prescribed
        precision.
    projection_type : str
        The type of projection. The default is 'interpolation'.

    '''

    def __init__(self):
        '''
        Constructor for the class FunctionalTensorPrincipalComponentAnalysis.

        Returns
        -------
        None.

        '''
        self.tol = 1e-8
        self.max_rank = np.inf
        self.bases = None
        self.grid = None
        self.display = True
        self.pca_sampling_factor = 1
        self.pca_adaptive_sampling = False
        self.projection_type = 'interpolation'

    def hopca(self, fun):
        '''
        Return the set of alpha-principal components of a tensor, for all alpha
        in {0,1,...,d-1}.

        For prescribed precision, set FPCA.max_rank = np.inf and FPCA.tol to
        the desired precision (possibly an array of length d).

        For prescribed rank, set FPCA.tol = np.inf and FPCA.max_rank to the
        desired rank (possibly an array of length d).

        See also the documentation of the class
        FunctionalTensorPrincipalComponentAnalysis.

        Parameters
        ----------
        fun : function
            A function of d variables.

        Raises
        ------
        ValueError
            If the provided tolerance and max ranks are not correct.

        Returns
        -------
        f_pc : list
            List of the alpha-principal components of the function.
        outputs : list
            List containing the outputs of the method
            alpha_principal_components.

        '''
        solver, t_fun, shape, tpca = self._prepare(fun)
        t_pc, outputs = tpca.hopca(t_fun, shape)

        P = solver._projection_operators()
        t_pc = [np.matmul(p, x) for p, x in zip(P, t_pc)]
        f_pc = [tensap.SubFunctionalBasis(b, a) for b, a in
                zip(solver.bases.bases, t_pc)]
        return f_pc, outputs

    def tucker_approximation(self, fun):
        '''
        Approximation of a function of d variables in Tucker format based on a
        Principal Component Analysis.

        For a prescribed precision, set TPCA.max_rank = np.inf and TPCA.tol to
        the desired precision (possibly an array of length d).

        For a prescribed rank, set TPCA.tol = np.inf and TPCA.max_rank to the
        desired rank (possibly an array of length d).

        See also the documentation of the class
        FunctionalTensorPrincipalComponentAnalysis.

        Parameters
        ----------
        fun : function
            A function of d variables.

        Raises
        ------
        ValueError
            If the provided tolerance and max ranks are not correct.

        Returns
        -------
        tensap.FunctionalTensor
            A function in tree based format with a trivial tree.
        dict
            Dictionnary containing the outputs of the method.

        '''
        solver, t_fun, shape, tpca = self._prepare(fun)
        tensor, output = tpca.tucker_approximation(t_fun, shape)
        return solver._project(tensor), output

    def tt_approximation(self, fun):
        '''
        Approximation of a function of d variables in tensor train format based
        on a Principal Component Analysis.

        For a prescribed precision, set TPCA.max_rank = np.inf and TPCA.tol to
        the desired precision (possibly an array of length d).

        For a prescribed rank, set TPCA.tol = np.inf and TPCA.max_rank to the
        desired rank (possibly an array of length d).

        See also the documentation of the class
        FunctionalTensorPrincipalComponentAnalysis.

        Parameters
        ----------
        fun : function
            A function of d variables.

        Raises
        ------
        ValueError
            If the provided tolerance and max ranks are not correct.

        Returns
        -------
        tensap.FunctionalTensor
            A function in tree based format with a linear tree.
        dict
            Dictionnary containing the outputs of the method.

        '''
        solver, t_fun, shape, tpca = self._prepare(fun)
        tensor, output = tpca.tt_approximation(t_fun, shape)
        return solver._project(tensor), output

    def tree_based_approximation(self, fun, tree, is_active_node=None):
        '''
        Approximation of a function of d variables in tree based tensor format
        based on a Principal Component Analysis.

        For a prescribed precision, set TPCA.max_rank = np.inf and TPCA.tol to
        the desired precision (possibly an array of length d-1).

        For a prescribed rank, set TPCA.tol = np.inf and TPCA.max_rank to the
        desired rank (possibly an array of length d-1).

        See also the documentation of the class
        FunctionalTensorPrincipalComponentAnalysis.

        Parameters
        ----------
        fun : fun or tensap.Function
            Function of d variables i_1, ..., i_d which returns the entries of
            the tensor.
        shape : list or numpy.ndarray
            The shape of the tensor.
        tree : tensap.DimensionTree
            The required dimension tree.
        is_active_node : list or numpy.ndarray, optional
            An array of booleans indicating which nodes of the tree are active.
            The default is None, settings all the nodes active.

        Raises
        ------
        ValueError
            If the provided tolerance and max ranks are not correct.

        Returns
        -------
        tensap.FunctionalTensor
            A function in tree based format.
        dict
            Dictionnary containing the outputs of the method.

        '''
        solver, t_fun, shape, tpca = self._prepare(fun)
        tensor, output = tpca.tree_based_approximation(t_fun, shape, tree,
                                                       is_active_node)
        return solver._project(tensor), output

    def _prepare(self, fun):
        '''
        Prepare the principal component analysis.

        Parameters
        ----------
        fun : function
            A function of d variables.

        Raises
        ------
        ValueError
            If the attribute projection_type is wrong.

        Returns
        -------
        tensap.FunctionalTensorPrincipalComponentAnalysis
            A FunctionalTensorPrincipalComponentAnalysis object with updated
            attributes.
        function
            A function used by the methods of
            tensap.TensorPrincipalComponentAnalysis.
        numpy.ndarray
            The cardinals of the functional bases.
        tensap.TensorPrincipalComponentAnalysis
            A tensap.TensorPrincipalComponentAnalysis used to perform the
            principal component analysis.

        '''
        solver = deepcopy(self)
        solver.bases = tensap.FunctionalBases(self.bases)

        # Create the tensor product grid
        if solver.projection_type == 'interpolation':
            if solver.grid is None:
                solver.grid = solver.bases.interpolation_points()
            else:
                solver.grid = solver.bases.interpolation_points(solver.grid)
        else:
            raise ValueError('Wrong projection_type attribute.')
        solver.grid = tensap.FullTensorGrid(solver.grid)

        # Create the function which provides the values of the function on the
        # grid
        def t_fun(i):
            return fun(solver.grid.eval_at_indices(i))

        shape = solver.bases.cardinals()

        # Create a TensorPrincipalComponentAnalysis with the same values of
        # properties as the FunctionalPrincipalComponentAnalysis
        tpca = tensap.TensorPrincipalComponentAnalysis()
        tpca.display = solver.display
        tpca.pca_sampling_factor = solver.pca_sampling_factor
        tpca.pca_adaptive_sampling = solver.pca_adaptive_sampling
        tpca.tol = solver.tol
        tpca.max_rank = solver.max_rank

        return solver, t_fun, shape, tpca

    def _projection_operators(self):
        '''
        Compute projection operators.

        The matrix P[nu] represents the operator which associates to the values
        of a function of the variable x_nu on a grid the coefficients of the
        projection on the basis of functions of the variable x_nu.

        Raises
        ------
        ValueError
            If the attribute projection_type is wrong.

        Returns
        -------
        P : list
            The projection operators for each dimension.

        '''
        if self.projection_type == 'interpolation':
            P = self.bases.eval(self.grid.grids)
            P = [np.linalg.pinv(x) for x in P]
        else:
            raise ValueError('Wrong projection_type attribute.')
        return P

    def _project(self, t):
        '''
        Compute the projection of a tensor on the functional bases.

        The method takes an AlgebraicTensor t whose entries are the values of
        the function on a product grid, and returns a FunctionalTensor obtained
        by applying the projections obtained by the method
        _projectionOperators.

        Parameters
        ----------
        t : tensap.Tensor
            The tensor used for the projection.

        Returns
        -------
        tensap.FunctionalTensor
            The obtained projection.

        '''
        tensor = deepcopy(t)
        P = self._projection_operators()
        for nu in range(tensor.order):
            alpha = tensor.tree.dim2ind[nu]
            if tensor.is_active_node[alpha-1]:
                data = np.matmul(P[nu], tensor.tensors[alpha-1].data)
                tensor.tensors[alpha-1] = tensap.FullTensor(data, 2, data.shape)
            else:
                pa = tensor.tree.parent(alpha)
                ch = tensor.tree.child_number(alpha)
                tensor.tensors[pa-1] = \
                    tensor.tensors[pa-1].tensor_matrix_product(P[nu], ch-1)
        return tensap.FunctionalTensor(tensor, self.bases)



# Copyright (c) 2020, Anthony Nouy, Erwan Grelier
# This file is part of tensap (tensor approximation package).

# tensap is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# tensap is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with tensap.  If not, see <https://www.gnu.org/licenses/>.

'''
Module linear_model_learning.

'''

import tensap


class LinearModelLearning(tensap.Learning):

    def __init__(self, loss):
        super().__init__(loss)
        self.basis = None
        self.basis_eval = None
        self.basis_eval_test = None
        self.regularization = False
        self.regularization_type = 'l1'
        self.regularization_options = {'alpha': 0}
        self.basis_adaptation = False
        self.basis_adaptation_path = None
        self.options = {}

    def initialize(self):
        # If the test error cannot be computed, it is disabled
        if self.test_error and self.basis is None and \
                self.basis_eval_test is None:
            print('The test error cannot be computed.')
            self.test_error = False

        # Bases evaluation
        try:
            if self.training_data is not None and self.basis_eval is None:
                if isinstance(self.training_data, list) and \
                        self.training_data[0] is not None:
                    self.basis_eval = self.basis.eval(self.training_data[0])
                elif not isinstance(self.training_data, list) and \
                        self.training_data is not None:
                    self.basis_eval = self.basis.eval(self.training_data)
                else:
                    raise ValueError('Must provide input training data.')

            if self.test_error and self.test_data is not None and \
                    self.basis_eval_test is None:
                if isinstance(self.test_data, list) and \
                        self.test_data[0] is not None:
                    self.basis_eval_test = self.basis.eval(
                        self.test_data[0])
                elif not isinstance(self.test_data, list) and \
                        self.test_data is not None:
                    self.basis_eval_test = self.basis.eval(self.test_data)
                else:
                    raise ValueError('Must provide input test data.')
        except Exception:
            pass



# Copyright (c) 2020, Anthony Nouy, Erwan Grelier
# This file is part of tensap (tensor approximation package).

# tensap is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# tensap is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with tensap.  If not, see <https://www.gnu.org/licenses/>.

'''
Module tensor_learning.

'''

from abc import abstractmethod
from copy import deepcopy
import pprint
import numpy as np
import tensap


class TensorLearning(tensap.Learning):
    '''
    Class TensorLearning.

    See also tensap.Learning.

    Attributes
    ----------
    order : int
        Order of the tensor.
    bases : list or numpy.ndarray
        The functional bases used for by the approximation.
    bases_eval : list or numpy.ndarray
        The evaluations of the bases on the training data.
    bases_eval_test : list or numpy.ndarray
        The evaluations of the bases on the test data.
    algorithm : str
        The choice of algorithm.
    initialization_type : str
        The type of initialization.
    initial_guess : tensap.Tensor
        An initial guess for the solver.
    tree_adaptation : bool
        Boolean indicating if tree adaptation is to be performed.
    tree_adaptation_options : dict
        Options for the tree adaptation.
    rank_adaptation : bool
        Boolean indicating if rank adaptation is to be performed.
    rank_adaptation_options : dict
        Options for the rank adaptation.
    tolerance : dict
        Tolerances for the solver.
    linear_model_learning_parameters : dict
        Parameters for the linear solvers.
    alternating_minimization_parameters : dict
        Parameters for the alternating minimization.
    bases_adaptation_path : list or numpy.ndarray
        Paths for each basis for the basis adaptation.
    store_iterates : bool
        Boolean indicating if the iterates are to be stored.
    rank : int or list or numpy.ndarray
        The rank of the expected approximation.
    output_dimension : int
        The dimension of the outputs of the function to be approximated.
    _number_of_parameters : int
        The number of parameters of the approximation.
    _exploration_strategy : numpy.ndarray
        The exploration strategy.
    _warnings : dict
        Dictionnary containing the state of some warnings to be displayed.

    '''

    def __init__(self, loss):
        '''
        Constructor for the class TensorLearning.

        See also tensap.Learning.

        Parameters
        ----------
        loss : tensap.LossFunction
            The loss function associated with the solver.

        Returns
        -------
        None.

        '''
        super().__init__(loss)
        self.linear_model_learning = tensap.Learning.linear_model(loss)
        self.order = None
        self.bases = None
        self.bases_eval = None
        self.bases_eval_test = None
        self.algorithm = 'standard'
        self.initialization_type = None
        self.initial_guess = None
        self.tree_adaptation = False
        self.tree_adaptation_options = {'tolerance': None,
                                        'max_iterations': 100,
                                        'force_rank_adaptation': True}
        self.rank_adaptation = False
        self.rank_adaptation_options = {'max_iterations': 10,
                                        'early_stopping': False,
                                        'early_stopping_factor': 10}
        self.tolerance = {'on_error': 1e-6, 'on_stagnation': 1e-6}
        self.linear_model_learning_parameters = \
            {'identical_for_all_parameters': True}
        self.alternating_minimization_parameters = {'display': False,
                                                    'max_iterations': 30,
                                                    'stagnation': 1e-6,
                                                    'random': False}
        self.bases_adaptation_path = None
        self.store_iterates = True
        self.rank = 1
        self.output_dimension = 1

        self.model_selection = False
        # Possible choices: 'test_error', 'cv_error'
        self.model_selection_options['type'] = 'test_error'

        self._number_of_parameters = None
        self._exploration_strategy = None
        self._warnings = {'orthonormality_warning_display': True,
                          'empty_bases_warning_display': True}

    def __repr__(self):
        return pprint.pformat(self.__dict__, indent=4, width=1)

    def solve(self, *args):
        '''
        Solver for the learning problem with tensor formats.

        Parameters
        ----------
        *args : misc
            Additional arguments, if needed.

        Raises
        ------
        NotImplementedError
            If the required solver is not implemented.
        ValueError
            If the required data is not provided.

        Returns
        -------
        f : tensap.FunctionalTensor
            The learned approximation.
        output : dict
            The outputs of the solver.

        '''
        # If possible, deduce from training_data the output dimension
        if isinstance(self.training_data, list) and \
                len(self.training_data) == 2 and \
                np.ndim(self.training_data[1]) == 2:
            self.output_dimension = self.training_data[1].shape[1]
        if self.output_dimension > 1 and \
            (not isinstance(self, tensap.TreeBasedTensorLearning) or
             not isinstance(self.loss_function, (tensap.SquareLossFunction,
                                                 tensap.CustomLossFunction))):
            raise NotImplementedError(
                'Solver not implemented for vector-valued functions ' +
                'approximation, use TreeBasedTensorLearning with a ' +
                'SquareLossFunction or a CustomLossFunction instead.')

        if self.output_dimension > 1 and \
            'type' in self.rank_adaptation_options and \
                self.rank_adaptation_options['type'] == 'inner':
            print('Inner rank adaptation not implemented for ' +
                  'output_dimension greater than 1, disabling it.')
            del self.rank_adaptation_options['type']

        if self._warnings['orthonormality_warning_display'] and \
            (self.bases is None or (self.bases is not None and not
                                    np.all([x.is_orthonormal for x in
                                            self.bases.bases]))):
            self._warnings['orthonormality_warning_display'] = False
            print('The implemented learning algorithms are designed ' +
                  'for orthonormal bases. These algorithms work with ' +
                  'non-orthonormal bases, but without some guarantees ' +
                  'on their results.')

        # If no bases are provided, warn that the returned functions are
        # evaluated on the training data
        if self.bases is None and \
                self._warnings['empty_bases_warning_display']:
            self._warnings['empty_bases_warning_display'] = False
            print('The returned functions are evaluated on the training ' +
                  'data. To evaluate them at other points, assign to the ' +
                  'FunctionalTensor a nonempty field bases and set the ' +
                  'attribute evaluatedBases to False.')

        # If the test error cannot be computed, it is disabled
        if self.test_error and \
            not isinstance(self.bases, tensap.FunctionalBases) and \
                self.bases_eval_test is None:
            print('The test error cannot be computed.')
            self.test_error = False

        # Assert if basis adaptation can be performed
        if self.bases_adaptation_path is None and \
                not hasattr(self.bases, 'adaptation_path'):
            if isinstance(self.linear_model_learning, (list, np.ndarray)) and \
                    np.any([x.basis_adaptation for
                            x in self.linear_model_learning]):
                print('Cannot perform basis adaptation, disabling it.')
                for linear_solver in self.linear_model_learning:
                    setattr(linear_solver, 'basis_adaptation', False)

            elif not isinstance(self.linear_model_learning,
                                (list, np.ndarray)) and \
                    self.linear_model_learning.basis_adaptation:
                print('Cannot perform basis adaptation, disabling it.')
                self.linear_model_learning.basis_adaptation = False

        # Bases evaluation
        if hasattr(self.bases, 'eval'):
            if self.training_data is not None and self.bases_eval is None:
                if isinstance(self.training_data, list) and \
                        self.training_data[0] is not None:
                    self.bases_eval = self.bases.eval(self.training_data[0])
                elif not isinstance(self.training_data, list) and \
                        self.training_data is not None:
                    self.bases_eval = self.bases.eval(self.training_data)
                else:
                    raise ValueError('Must provide input training data.')

            if self.test_error and self.test_data is not None and \
                    self.bases_eval_test is None:
                if isinstance(self.test_data, list) and \
                        self.test_data[0] is not None:
                    self.bases_eval_test = self.bases.eval(self.test_data[0])
                elif not isinstance(self.test_data, list) and \
                        self.test_data is not None:
                    self.bases_eval_test = self.bases.eval(self.test_data)
                else:
                    raise ValueError('Must provide input test data.')

        if self.model_selection:
            self.store_iterates = True
            if self.model_selection_options['type'] == 'test_error' and \
                    not self.test_error:
                print('Cannot perform test error based model selection, ' +
                      'disabling it.')
                self.model_selection = False
            elif self.model_selection_options['type'] == 'cv_error' and \
                    not self.error_estimation:
                self.error_estimation = True

        if self.rank_adaptation:
            if 'type' not in self.rank_adaptation_options or \
                    self.rank_adaptation_options['type'] == 'standard':
                fun, output = self._solve_adaptation()
            elif isinstance(self.rank_adaptation_options['type'], str):
                adapt_type = self.rank_adaptation_options['type']
                if adapt_type == 'dmrg' or adapt_type == 'dmrg_low_rank':
                    adapt_type = 'dmrg'

                # Call the method corresponding to the rank adaptation option
                fun, output = eval('self._solve_' + adapt_type.lower() +
                                   '_rank_adaptation()')
            else:
                raise ValueError('The rank_adaptation_options attribute ' +
                                 'must be either inexistant or a string.')
        elif self.algorithm == 'standard':
            fun, output = self._solve_standard()
        else:
            fun, output = eval('self._solve_' + self.algorithm.lower() + '()')

        if self.model_selection:
            if self.model_selection_options['type'] == 'test_error':
                ind = np.argmin(output['test_error_iterations'])
                fun = output['iterates'][ind]
                output['selected_model_number'] = ind
                if 'error_iterations' in output and \
                    np.size(output['error_iterations']) > 0:
                    output['error'] = output['error_iterations'][ind]
                output['test_error'] = output['test_error_iterations'][ind]
                if self.display:
                    print('\nModel selection using the test error: model ' +
                          '#%i selected' % ind)
            elif self.model_selection_options['type'] == 'cv_error':
                ind = np.argmin(output['error_iterations'])
                fun = output['iterates'][ind]
                output['selected_model_number'] = ind
                if 'test_error_iterations' in output and \
                    np.size(output['test_error_iterations']) > 0:
                    output['test_error'] = output['test_error_iterations'][ind]
                output['error'] = output['error_iterations'][ind]
                if self.display:
                    print('\nModel selection using the cross-validation ' +
                          ' error: model #%i selected' % ind)
            else:
                print('Wrong model selection type, returning the last ' +
                      'iterate.')

            if self.display:
                if self.alternating_minimization_parameters['display']:
                    print('')
                self.final_display(fun)
                if 'error' in output:
                    print(', error = %2.5e' % output['error'], end='')
                if 'test_error' in output:
                    print(', test error = %2.5e' % output['test_error'],
                          end='')
                print('')

        return fun, output

    def _solve_standard(self):
        '''
        Solver for the learning problem with tensor formats using the standard
        algorithm (without adaptation).

        Raises
        ------
        ValueError
            If the number of LinearModelLearning objects is not equal to
            _numberOfParameters.

        Returns
        -------
        f : tensap.FunctionalTensor
            The learned approximation.
        output : dict
            The outputs of the solver.

        '''
        output = {'flag': 0}

        self, f = self.initialize()
        f = tensap.FunctionalTensor(f, self.bases_eval)

        # Replication of the LinearModelLearning objects
        if self.linear_model_learning_parameters[
                'identical_for_all_parameters'] and \
                not isinstance(self.linear_model_learning, (list, np.ndarray)):
            self.linear_model_learning = list(map(deepcopy, [
                self.linear_model_learning] * self._number_of_parameters))
        elif isinstance(self.linear_model_learning, (list, np.ndarray)) and \
                len(self.linear_model_learning) != self._number_of_parameters:
            raise ValueError('Must provide self._numberOfParameters ' +
                             'LinearModelLearning objects.')

        if self.error_estimation:
            for x in self.linear_model_learning:
                setattr(x, 'error_estimation', True)

        # Working set paths
        if isinstance(self.linear_model_learning, (list, np.ndarray)) and \
            np.any([x.basis_adaptation for
                    x in self.linear_model_learning]) and \
                self.bases_adaptation_path is None:
            self.bases_adaptation_path = self.bases.adaptation_path()

        if self.alternating_minimization_parameters['max_iterations'] == 0:
            return f, output

        output['stagnation_criterion'] = []
        output['iterates'] = []
        output['error_iterations'] = []
        output['test_error_iterations'] = []

        # Alternating minimization loop
        for iteration in range(self.alternating_minimization_parameters
                               ['max_iterations']):
            self, f = self.pre_processing(f)
            f0 = deepcopy(f)

            if self.alternating_minimization_parameters['random']:
                # Randomize the exploration strategy
                alpha_list = self.randomize_exploration_strategy()
            else:
                alpha_list = self._exploration_strategy

            for alpha in alpha_list:
                self, A, b, f = \
                    self.prepare_alternating_minimization_system(f, alpha)
                self.linear_model_learning[alpha-1].training_data = [None, b]
                self.linear_model_learning[alpha-1].basis = None
                self.linear_model_learning[alpha-1].basis_eval = A

                coef, output_tmp = self.linear_model_learning[alpha-1].solve()
                if coef is None or np.count_nonzero(coef) == 0 or \
                        not np.all(np.isfinite(coef)):
                    print('Empty, zero or NaN solution, returning to the ' +
                          'previous iteration.')
                    output['flag'] = -2
                    output['error'] = np.inf
                    break

                f = self.set_parameter(f, alpha, coef)

            stagnation = self.stagnation_criterion(f, f0)
            output['stagnation_criterion'].append(stagnation)

            if self.store_iterates:
                if isinstance(self.bases, tensap.FunctionalBases):
                    output['iterates'].append(tensap.FunctionalTensor(
                        f.tensor, self.bases))
                else:
                    output['iterates'].append(f)

            if 'error' in output_tmp:
                output['error'] = output_tmp['error']
                output['error_iterations'].append(output['error'])

            if self.test_error:
                f_eval_test = tensap.FunctionalTensor(f, self.bases_eval_test)
                output['test_error'] = self.loss_function.test_error(
                    f_eval_test, self.test_data)
                output['test_error_iterations'].append(output['test_error'])

            if self.alternating_minimization_parameters['display']:
                print('\tAlt. min. iteration %s: stagnation = %2.5e' %
                      (str(iteration).
                       zfill(len(str(self.alternating_minimization_parameters
                                     ['max_iterations']-1))),
                       stagnation), end='')
                if 'error' in output:
                    print(', error = %2.5e' % output['error'], end='')
                if self.test_error:
                    if not np.isscalar(output['test_error']):
                        output['test_error'] = output['test_error'].numpy()
                    print(', test error = %2.5e' % output['test_error'],
                          end='')
                print('')

            if iteration > 0 and stagnation < \
                    self.alternating_minimization_parameters['stagnation']:
                output['flag'] = 1
                break

            if self.test_error and \
                    output['test_error'] < self.tolerance['on_error']:
                output['flag'] = 1
                break

        if isinstance(self.bases, tensap.FunctionalBases):
            f = tensap.FunctionalTensor(f.tensor, self.bases)
        output['iter'] = iteration

        if self.display and not self.model_selection:
            if self.alternating_minimization_parameters['display']:
                print('')
            self.final_display(f)
            if 'error' in output:
                print(', error = %2.5e' % output['error'], end='')
            if 'test_error' in output:
                print(', test error = %2.5e' % output['test_error'], end='')
            print('')

        return f, output

    def _solve_adaptation(self):
        '''
        Solver for the learning problem with tensor formats using the adaptive
        algorithm.

        Returns
        -------
        f : tensap.FunctionalTensor
            The learned approximation.
        output : dict
            The outputs of the solver.

        '''

        flag = 0
        output = {'enriched_nodes_iterations':
                  np.empty(self.rank_adaptation_options['max_iterations'],
                           dtype=object)}
        tree_adapt = False

        f = None
        errors = np.zeros(self.rank_adaptation_options['max_iterations'])
        test_errors = np.zeros(self.rank_adaptation_options['max_iterations'])
        iterates = np.empty(self.rank_adaptation_options['max_iterations'],
                            dtype=object)

        # new_rank = s_local.rank
        new_rank = self.rank
        s_local = self.local_solver()
        s_local.model_selection = False

        enriched_nodes = np.array([])
        for iteration in range(self.rank_adaptation_options['max_iterations']):
            s_local.bases = self.bases
            s_local.bases_eval = self.bases_eval
            s_local.bases_eval_test = self.bases_eval_test
            s_local.training_data = self.training_data
            s_local.test_data = self.test_data
            s_local.rank = new_rank

            f_old = deepcopy(f)
            f, output_local = s_local.solve()
            s_local = self.local_solver()

            if 'error' in output_local:
                errors[iteration] = output_local['error']
                if np.isinf(errors[iteration]):
                    print('Infinite error, returning to the previous iterate.')
                    f = f_old
                    iteration -= 1
                    flag = -2
                    break

            if self.test_error:
                f_eval_test = tensap.FunctionalTensor(
                    f, self.bases_eval_test)
                test_errors[iteration] = self.loss_function.test_error(
                    f_eval_test, self.test_data)

            if self.store_iterates:
                if isinstance(self.bases, tensap.FunctionalBases):
                    iterates[iteration] = tensap.FunctionalTensor(f.tensor,
                                                                  self.bases)
                else:
                    iterates[iteration] = f

            if self.display:
                if self.alternating_minimization_parameters['display']:
                    print('')
                print('\nRank adaptation, iteration %i:' % (iteration))
                self.adaptation_display(f, enriched_nodes)
                print('\tStorage complexity = %i' % f.tensor.storage(), flush=True)

                if errors[iteration] != 0:
                    print('\tError      = %2.5e' % errors[iteration])
                if test_errors[iteration] != 0:
                    print('\tTest error = %2.5e' % test_errors[iteration])
                if self.alternating_minimization_parameters['display']:
                    print('')

            if iteration == self.rank_adaptation_options['max_iterations']-1:
                break

            if (self.test_error and test_errors[iteration] <
                    self.tolerance['on_error']) or \
                ('error' in output_local and errors[iteration] <
                 self.tolerance['on_error']):
                flag = 1
                break

            fac = self.rank_adaptation_options['early_stopping_factor']
            cond = iteration > 0 and (self.test_error and
                                      (np.isnan(test_errors[iteration]) or
                                       fac * np.min(test_errors[:iteration]) <
                                       test_errors[iteration]) or
                                      ('error' in output_local and
                                       (np.isnan(errors[iteration]) or
                                        fac * np.min(errors[:iteration]) <
                                        errors[iteration])))
            if self.rank_adaptation_options['early_stopping'] and cond:
                print('Early stopping', end='')
                if 'error' in output_local:
                    print(', error = %2.5e' % errors[iteration], end='')
                if self.test_error:
                    print(', test error = %2.5e' % test_errors[iteration],
                          end='')
                print('\n')
                iteration -= 1
                f = f_old
                flag = -1
                break

            adapted_tree = False
            if s_local.tree_adaptation and iteration > 0 and \
                    (not self.tree_adaptation_options['force_rank_adaptation']
                     or not tree_adapt):
                C_old = f.tensor.storage()
                self, f, output = self.adapt_tree(f, errors[iteration],
                                                  None, output, iteration)
                adapted_tree = output['adapted_tree']
                if adapted_tree:
                    if self.display:
                        print('\t\tStorage complexity before permutation ' +
                              '= %i' % C_old)
                        print('\t\tStorage complexity after permutation ' +
                              '= %i' % f.tensor.storage())
                    if self.test_error:
                        f_eval_test = tensap.FunctionalTensor(
                            f, self.bases_eval_test)
                        test_errors[iteration] = self.loss_function.test_error(
                            f_eval_test, self.test_data)
                        if self.display:
                            print('\t\tTest error after permutation ' +
                                  '= %2.5e' % test_errors[iteration])

                    if self.alternating_minimization_parameters['display']:
                        print('')

            if not self.tree_adaptation or not adapted_tree:
                if iteration > 0 and not tree_adapt:
                    stagnation = self.stagnation_criterion(
                        tensap.FunctionalTensor(f.tensor, self.bases_eval),
                        tensap.FunctionalTensor(f_old.tensor, self.bases_eval))
                    if stagnation < self.tolerance['on_stagnation'] or \
                            np.isnan(stagnation):
                        break
                tree_adapt = False
                new_rank, enriched_nodes, tensor_for_initialization = \
                    self.new_rank_selection(f)
                output['enriched_nodes_iterations'][iteration] = enriched_nodes
                s_local = self.initial_guess_new_rank(
                    s_local, tensor_for_initialization, new_rank)
            else:
                tree_adapt = True
                enriched_nodes = []
                new_rank = f.tensor.ranks
                s_local.initialization_type = 'initial_guess'
                s_local.initial_guess = f.tensor

        if isinstance(self.bases, tensap.FunctionalBases):
            f = tensap.FunctionalTensor(f.tensor, self.bases)

        if self.store_iterates:
            output['iterates'] = iterates[:iteration+1]

        output['flag'] = flag
        output['enriched_nodes_iterations'] = \
            output['enriched_nodes_iterations'][:iteration+1]
        if 'error' in output_local:
            output['error_iterations'] = errors[:iteration+1]
            output['error'] = errors[iteration]

        if self.test_error:
            output['test_error_iterations'] = test_errors[:iteration+1]
            output['test_error'] = test_errors[iteration]

        if 'adapted_tree' in output:
            del output['adapted_tree']

        return f, output

    def adapt_tree(self, f, cv_error, test_error, output, *args):
        '''
        Tree adaptation algorithm.

        Parameters
        ----------
        f : tensap.Tensor
            The current tensor approximation.
        cv_error : float
            The current cross-validation error.
        test_error : float
            The current test error.
        output : dict
            The outputs of the solver.
        args : tuple
            Additional parameters.

        Returns
        -------
        tensap.TensorLearning
            The TensorLearning object.
        f : tensap.Tensor
            The current tensor approximation, perhaps with an adapted tree.
        output : dict
            The outputs of the solver.

        '''
        return self, f, output

    @abstractmethod
    def initialize(self):
        '''
        Initialization of the learning algorithm.

        Returns
        -------
        tensap.TensorLearning
            The TensorLearning object.
        f : tensap.FunctionalTensor
            The initialization of the solver.

        '''

    @abstractmethod
    def pre_processing(self, f):
        '''
        Initialization of the alternating minimization algorithm.

        Parameters
        ----------
        f : tensap.Tensor
            The current tensor approximation.

        Returns
        -------
        tensap.TensorLearning
            The TensorLearning object.
        f : tensap.Tensor
            The pre-processed tensor approximation.

        '''

    @abstractmethod
    def randomize_exploration_strategy(self):
        '''
        Randomization of the exploration strategy.

        Returns
        -------
        alpha_list : numpy.ndarray
            The randomized exploration strategy.

        '''

    @abstractmethod
    def prepare_alternating_minimization_system(self, f, mu):
        '''
        Preparation of the alternating minimization algorithm.

        Parameters
        ----------
        f : tensap.FunctionalTensor
            The current approximation.
        mu : int
            The number of the parameter to be optimized.

        Returns
        -------
        tensap.TensorLearning
            The TensorLearning object.
        A : numpy.ndarray
            The matricized partial evaluation of f.
        b : numpy.ndarray
            The target training data.
        f : tensap.FunctionalTensor
            The current approximation.

        '''

    @abstractmethod
    def set_parameter(self, f, mu, coef):
        '''
        Update of the parameter mu of the tensor f.

        Parameters
        ----------
        f : tensap.FunctionalTensor
            The current approximation.
        mu : int
            The number of the optimized parameter.
        coef : numpy.ndarray
            The new value of the parameter mu of f.

        Returns
        -------
        f : tensap.FunctionalTensor
            The updated approximation.

        '''

    @abstractmethod
    def stagnation_criterion(self, f, f0):
        '''
        Computation of the stagnation criterion.

        Parameters
        ----------
        f : tensap.FunctionalTensor
            The current approximation.
        f0 : tensap.FunctionalTensor
            The previous approximation.

        Returns
        -------
        stagnation : float
            The value of the stagnation criterion.

        '''

    @abstractmethod
    def final_display(self, f):
        '''
        Display at the end of the computation.

        Parameters
        ----------
        f : tensap.FunctionalTensor
            The current approximation.

        Returns
        -------
        None.

        '''

    @abstractmethod
    def local_solver(self):
        '''
         Extraction of the solver for the adaptive algorithm

        Returns
        -------
        s_local : tensap.TensorLearning
            The local solver.

        '''

    @abstractmethod
    def new_rank_selection(self, f):
        '''
        Selection of a new rank in the adaptive algorithm.

        Parameters
        ----------
        f : tensap.FunctionalTensor
            The current approximation.

        Returns
        -------
        new_rank : numpy.ndarray
            The new tensor rank.
        enriched_nodes : numpy.ndarray
            The enriched parameters.
        tensor_for_initialization : tensap.Tensor
            A tensor used for the initialization of the next iteration.

        '''

    @abstractmethod
    def initial_guess_new_rank(self, s_local, f, new_rank):
        '''
        Computation of the initial guess with the new selected rank.

        Parameters
        ----------
        s_local : tensap.TensorLearning
            The local solver.
        f : tensap.FunctionalTensor
            The current approximation.
        new_rank : numpy.ndarray
            The new tensor rank.

        Returns
        -------
        s_local : tensap.TensorLearning
            The local solver with the initial guess in s_local.initial_guess.

        '''

    @abstractmethod
    def adaptation_display(self, f, enriched_nodes):
        '''
        Display during the adaptation.

        Parameters
        ----------
        f : tensap.FunctionalTensor
            The current approximation.
        enriched_nodes : numpy.ndarray
            The enriched parameters.
        '''
        pass



# Copyright (c) 2020, Anthony Nouy, Erwan Grelier
# This file is part of tensap (tensor approximation package).

# tensap is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# tensap is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with tensap.  If not, see <https://www.gnu.org/licenses/>.

'''
Module chebyshev_points.

'''

import numpy as np


def chebyshev_points(n, s=None):
    '''
    Return the first n Chebyshev points in [s[0], s[1]].

    Parameters
    ----------
    n : int
        The number of Chebyshev points.
    s : list or numpy.ndarray, optional
        The interval on which the points are computed. The default is [-1, 1].

    Returns
    -------
    x : numpy.ndarray
        The first n Chebyshev points in [s[0], s[1]].

    '''
    x = np.cos(np.pi*(2*np.arange(1, n+1)-1)/2/n)
    if s is not None:
        x = 0.5*(s[0]+s[1]) + 0.5*(s[1]-s[0])*x
    return np.expand_dims(x, 1)



# Copyright (c) 2020, Anthony Nouy, Erwan Grelier
# This file is part of tensap (tensor approximation package).

# tensap is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# tensap is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with tensap.  If not, see <https://www.gnu.org/licenses/>.

'''
Tutorial on tensor interpolation.

'''

import numpy as np
import tensap

# %% Definitions
D = 3
P = 10

F = tensap.UserDefinedFunction('1+x0+(2+x0)/(2+x1)+0.04*(x2-x2**3)', D)
F.evaluation_at_multiple_points = True
V = tensap.UniformRandomVariable(-1, 1)
X = tensap.RandomVector(V, D)
BASIS = tensap.PolynomialFunctionalBasis(tensap.LegendrePolynomials(),
                                         range(P+1))
BASES = tensap.FunctionalBases.duplicate(BASIS, D)

# %% Sparse tensor product functional basis
IND = tensap.MultiIndices.with_bounded_norm(D, 1, P)
H = tensap.SparseTensorProductFunctionalBasis(BASES, IND)

# %% Interpolation on a magic grid
G, _, _ = H.magic_points(X.random(1000))
IF = H.interpolate(F, G)

X_TEST = X.random(1000)
ERR = np.linalg.norm(F(X_TEST)-IF(X_TEST)) / np.linalg.norm(F(X_TEST))
print('Test error = %2.5e' % ERR)

# %% Interpolation on a structured magic grid
GRIDS = X.random(1000)
G, _ = BASES.magic_points([GRIDS[:, i] for i in range(GRIDS.shape[1])])
G = tensap.SparseTensorGrid(G, H.indices)
IF = H.interpolate(F, G.array())

X_TEST = X.random(1000)
ERR = np.linalg.norm(F(X_TEST)-IF(X_TEST)) / np.linalg.norm(F(X_TEST))
print('Test error = %2.5e' % ERR)

# %% Interpolation on a structured magic grid (alternative)
GRIDS = X.random(1000)
G, _ = BASES.magic_points([GRIDS[:, i] for i in range(GRIDS.shape[1])])
IF, OUTPUT = H.tensor_product_interpolation(F, G)

X_TEST = X.random(1000)
ERR = np.linalg.norm(F(X_TEST)-IF(X_TEST)) / np.linalg.norm(F(X_TEST))
print('Test error = %2.5e' % ERR)
